class CapturedException(Exception):
    pass


class AnalysisException(CapturedException):
    """
    Failed to analyze a SQL query plan.
    """


class ParseException(CapturedException):
    """
    Failed to parse a SQL command.
    """


class IllegalArgumentException(CapturedException):
    """
    Failed to analyze a SQL query plan.
    """


class StreamingQueryException(CapturedException):
    """
    Exception that stopped a :class:`StreamingQuery`.
    """


class QueryExecutionException(CapturedException):
    """
    Failed to execute a query.
    """


class PythonException(CapturedException):
    """
    Exceptions thrown from Python workers.
    """


class UnknownException(CapturedException):
    """
    None of the above exceptions.
    """


def convert_exception(e):
    """In pyspark, this method converts a received string into a Python exception."""
    return


def capture_sql_exception(f):
    """This is a decorator to convert py4j exceptions into nicer Python exceptions."""
    return


def install_exception_handler():
    """Don't implement, but also don't raise."""
    return


def toJArray(gateway, jtype, arr):
    """Convert python list to java type array"""
    return


def require_test_compiled():
    """ Raise Exception if test classes are not compiled"""

    raise RuntimeError("Tests don't exist for gelanis as they do for Spark.")


class ForeachBatchFunction:
    """
    This is the Python implementation of Java interface 'ForeachBatchFunction'. This wraps
    the user-defined 'foreachBatch' function such that it can be called from the JVM when
    the query is active.
    """
    # ==> Not implemented; do we actually need it?
    # def __init__(self, sql_ctx, func):
    #     self.sql_ctx = sql_ctx
    #     self.func = func
    #
    # def call(self, jdf, batch_id):
    #     from pyspark.sql.dataframe import DataFrame
    #     try:
    #         self.func(DataFrame(jdf, self.sql_ctx), batch_id)
    #     except Exception as e:
    #         self.error = e
    #         raise e
    #
    # class Java:
    #     implements = ['org.apache.spark.sql.execution.streaming.sources.PythonForeachBatchFunction']


def to_str(value):
    """
    A wrapper over str(), but converts bool values to lower case strings.
    If None is given, just returns None, instead of converting it to string "None".
    """
    if isinstance(value, bool):
        return str(value).lower()

    if value is None:
        return value

    return str(value)



from functools import partial
import itertools
import json

from ..._casts import get_struct_caster
from ..._row import create_row, row_from_keyed_values
from ..._schema_utils import infer_schema_from_rdd
from ...internal_utils.options import Options
from ...internal_utils.readers.utils import get_records, resolve_partitions
from ...types import StructType


class JSONReader:
    default_options = dict(
        primitivesAsString=False,
        prefersDecimal=False,
        allowComments=False,
        allowUnquotedFieldNames=False,
        allowSingleQuotes=True,
        allowNumericLeadingZero=False,
        allowBackslashEscapingAnyCharacter=False,
        mode="PERMISSIVE",
        columnNameOfCorruptRecord="",
        dateFormat="yyyy-MM-dd",
        timestampFormat="yyyy-MM-dd'T'HH:mm:ss.SSSXXX",
        multiLine=False,
        allowUnquotedControlChars=False,
        encoding=None,
        lineSep=None,
        samplingRatio=1.0,
        dropFieldIfAllNull=False,
        locale="en-US",
    )

    def __init__(self, spark, paths, schema, options):
        self.spark = spark
        self.paths = paths
        self.schema = schema
        self.options = Options(self.default_options, options)

    def read(self):
        sc = self.spark._sc
        paths = self.paths

        partitions, partition_schema = resolve_partitions(paths)

        rdd_filenames = sc.parallelize(sorted(partitions.keys()), len(partitions))
        rdd = rdd_filenames.flatMap(partial(
            parse_json_file,
            partitions,
            partition_schema,
            self.schema,
            self.options
        ))

        inferred_schema = infer_schema_from_rdd(rdd)

        schema = self.schema if self.schema is not None else inferred_schema
        schema_fields = {
            field.name: field
            for field in schema.fields
        }

        # Field order is defined by fields in the record, not by the given schema
        # Field type is defined by the given schema or inferred
        full_schema = StructType(
            fields=[
                schema_fields.get(field.name, field)
                for field in inferred_schema.fields
            ]
        )

        cast_row = get_struct_caster(inferred_schema, full_schema, options=self.options)
        casted_rdd = rdd.map(cast_row)
        casted_rdd._name = paths

        # pylint: disable=import-outside-toplevel, cyclic-import
        from ...internals import DataFrameInternal

        return DataFrameInternal(
            sc,
            casted_rdd,
            schema=full_schema
        )


def parse_json_file(partitions, partition_schema, schema, options, file_name):
    records = get_records(file_name, options.linesep, options.encoding)
    rows = []
    for record in records:
        partition = partitions[file_name]
        row = parse_record(record, schema, partition, partition_schema, options)
        row.set_input_file_name(file_name)
        rows.append(row)
    return rows


def parse_record(record, schema, partition, partition_schema, options):
    raw_record_value = json.loads(record, encoding=options.encoding)
    if not isinstance(raw_record_value, dict):
        raise NotImplementedError(
            "Top level items should be JSON objects (dicts),"
            f" got {type(raw_record_value)} with {raw_record_value}"
        )
    record_value = decode_record(raw_record_value)
    if schema is not None:
        record_fields = record_value.__fields__
        available_names = tuple(partition_schema.names) + record_fields
        field_names = [name for name in record_fields if name in schema.names] + [
            f.name for f in schema.fields if f.name not in available_names
        ]
    else:
        field_names = list(record_value.__fields__)
    record_values = [
        record_value[field_name] if field_name in record_value.__fields__ else None
        for field_name in field_names
    ]
    partition_field_names = [f.name for f in partition_schema.fields] if partition_schema else []
    # pylint: disable=W0511
    # todo: handle nested rows
    row = create_row(
        itertools.chain(field_names, partition_field_names),
        itertools.chain(record_values, partition)
    )
    return row


def decode_record(item):
    if isinstance(item, list):
        return [decode_record(e) for e in item]
    if isinstance(item, dict):
        return row_from_keyed_values(
            (key, decode_record(value))
            for key, value in item.items()
        )
    return item



from . import BinaryOperation, Expression, UnaryExpression
from ..utils import AnalysisException


class ArraysOverlap(BinaryOperation):
    pretty_name = "array_overlap"

    def eval(self, row, schema):
        set1 = set(self.arg1.eval(row, schema))
        set2 = set(self.arg2.eval(row, schema))
        overlap = set1 & set2
        if len(overlap) > 1 or (len(overlap) == 1 and None not in overlap):
            return True
        if set1 and set2 and (None in set1 or None in set2):
            return None
        return False


class ArrayContains(Expression):
    pretty_name = "array_contains"

    def __init__(self, array, value):
        self.array = array
        self.value = value.get_literal_value()
        super().__init__(array)

    def eval(self, row, schema):
        array_eval = self.array.eval(row, schema)
        if array_eval is None:
            return None
        return self.value in array_eval

    def args(self):
        return (
            self.array,
            self.value
        )


class ArrayColumn(Expression):
    pretty_name = "array"

    def __init__(self, *columns):
        super().__init__(columns)
        self.columns = columns

    def eval(self, row, schema):
        return [col.eval(row, schema) for col in self.columns]

    def args(self):
        return self.columns


class MapColumn(Expression):
    pretty_name = "map"

    def __init__(self, *columns):
        super().__init__(columns)
        self.columns = columns
        self.keys = columns[::2]
        self.values = columns[1::2]

    def eval(self, row, schema):
        return dict(
            (key.eval(row, schema), value.eval(row, schema))
            for key, value in zip(self.keys, self.values)
        )

    def args(self):
        return self.columns


class MapFromArraysColumn(Expression):
    pretty_name = "map_from_arrays"

    def __init__(self, keys, values):
        super().__init__(keys, values)
        self.keys = keys
        self.values = values

    def eval(self, row, schema):
        return dict(
            zip(self.keys.eval(row, schema), self.values.eval(row, schema))
        )

    def args(self):
        return (
            self.keys,
            self.values
        )


class Size(UnaryExpression):
    pretty_name = "size"

    def eval(self, row, schema):
        column_value = self.column.eval(row, schema)
        if isinstance(column_value, (list, dict)):
            return len(column_value)
        raise AnalysisException(
            f"{self.column} value should be an array or a map, got {type(column_value)}"
        )


class ArraySort(UnaryExpression):
    pretty_name = "array_sort"

    def eval(self, row, schema):
        return sorted(self.column.eval(row, schema))


class ArrayMin(UnaryExpression):
    pretty_name = "array_min"

    def eval(self, row, schema):
        return min(self.column.eval(row, schema))


class ArrayMax(UnaryExpression):
    pretty_name = "array_max"

    def eval(self, row, schema):
        return max(self.column.eval(row, schema))


class Slice(Expression):
    pretty_name = "slice"

    def __init__(self, column, start, length):
        self.column = column
        self.start = start.get_literal_value()
        self.length = length.get_literal_value()
        super().__init__(column)

    def eval(self, row, schema):
        return self.column.eval(row, schema)[self.start, self.start + self.length]

    def args(self):
        return (
            self.column,
            self.start,
            self.length
        )


class ArrayRepeat(Expression):
    pretty_name = "array_repeat"

    def __init__(self, col, count):
        super().__init__(col)
        self.col = col
        self.count = count.get_literal_value()

    def eval(self, row, schema):
        value = self.col.eval(row, schema)
        return [value for _ in range(self.count)]

    def args(self):
        return (
            self.col,
            self.count
        )


class Sequence(Expression):
    pretty_name = "array_join"

    def __init__(self, start, stop, step):
        super().__init__(start, stop, step)
        self.start = start
        self.stop = stop
        self.step = step

    def eval(self, row, schema):
        start_value = self.start.eval(row, schema)
        stop_value = self.stop.eval(row, schema)
        if self.step is not None:
            step_value = self.step.eval(row, schema)
            if (step_value < stop_value and step_value <= 0) \
               or (step_value > stop_value and step_value >= 0):
                raise Exception(
                    "requirement failed: Illegal sequence boundaries:"
                    f" {start_value} to {stop_value} by {step_value}"
                )
        else:
            step_value = 1 if start_value < stop_value else -1

        return list(range(start_value, stop_value, step_value))

    def args(self):
        if self.step is None:
            # Spark use the same logic of not displaying step
            # if it is None, even if it was explicitly set
            return (
                self.start,
                self.stop
            )
        return (
            self.start,
            self.stop,
            self.step
        )


class ArrayJoin(Expression):
    pretty_name = "array_join"

    def __init__(self, column, delimiter, nullReplacement):
        super().__init__(column)
        self.column = column
        self.delimiter = delimiter.get_literal_value()
        self.nullReplacement = nullReplacement.get_literal_value()

    def eval(self, row, schema):
        column_eval = self.column.eval(row, schema)
        return self.delimiter.join(
            value if value is not None else self.nullReplacement for value in column_eval
        )

    def args(self):
        if self.nullReplacement is None:
            # Spark use the same logic of not displaying nullReplacement
            # if it is None, even if it was explicitly set
            return (
                self.column,
                self.delimiter
            )
        return (
            self.column,
            self.delimiter,
            self.nullReplacement
        )


class SortArray(Expression):
    pretty_name = "sort_array"

    def __init__(self, col, asc):
        super().__init__(col)
        self.col = col
        self.asc = asc.get_literal_value()

    def eval(self, row, schema):
        return sorted(self.col.eval(row, schema), reverse=not self.asc)

    def args(self):
        return (
            self.col,
            self.asc
        )


class ArraysZip(Expression):
    pretty_name = "arrays_zip"

    def __init__(self, columns):
        super().__init__(*columns)
        self.columns = columns

    def eval(self, row, schema):
        return [
            list(combination)
            for combination in zip(
                *(c.eval(row, schema) for c in self.columns)
            )
        ]

    def args(self):
        return self.columns


class Flatten(UnaryExpression):
    pretty_name = "flatten"

    def eval(self, row, schema):
        return [
            value
            for array in self.column.eval(row, schema)
            for value in array
        ]


class ArrayPosition(Expression):
    pretty_name = "array_position"

    def __init__(self, col, value):
        super().__init__(col)
        self.col = col
        self.value = value.get_literal_value()

    def eval(self, row, schema):
        if self.value is None:
            return None
        col_eval = self.col.eval(row, schema)
        if col_eval is None:
            return None

        return col_eval.find(self.value) + 1

    def args(self):
        return (
            self.col,
            self.value
        )


class ElementAt(Expression):
    pretty_name = "element_at"

    def __init__(self, col, extraction):
        super().__init__(col)
        self.col = col
        self.extraction = extraction.get_literal_value()

    def eval(self, row, schema):
        col_eval = self.col.eval(row, schema)
        if isinstance(col_eval, list):
            return col_eval[self.extraction - 1]
        return col_eval.get(self.extraction)

    def args(self):
        return (
            self.col,
            self.extraction
        )


class ArrayRemove(Expression):
    pretty_name = "array_remove"

    def __init__(self, col, element):
        super().__init__(col, element)
        self.col = col
        self.element = element.get_literal_value()

    def eval(self, row, schema):
        array = self.col.eval(row, schema)
        return [value for value in array if value != self.element]

    def args(self):
        return (
            self.col,
            self.element
        )


class ArrayDistinct(UnaryExpression):
    pretty_name = "array_distinct"

    def eval(self, row, schema):
        return list(set(self.column.eval(row, schema)))


class ArrayIntersect(BinaryOperation):
    pretty_name = "array_intersect"

    def eval(self, row, schema):
        return list(set(self.arg1.eval(row, schema)) & set(self.arg2.eval(row, schema)))


class ArrayUnion(BinaryOperation):
    pretty_name = "array_union"

    def eval(self, row, schema):
        return list(set(self.arg1.eval(row, schema)) | set(self.arg2.eval(row, schema)))


class ArrayExcept(BinaryOperation):
    pretty_name = "array_except"

    def eval(self, row, schema):
        return list(set(self.arg1.eval(row, schema)) - set(self.arg2.eval(row, schema)))


__all__ = [
    "ArraysZip", "ArrayRepeat", "Flatten", "ArrayMax", "ArrayMin", "SortArray", "Size",
    "ArrayExcept", "ArrayUnion", "ArrayIntersect", "ArrayDistinct", "ArrayRemove", "ArraySort",
    "ElementAt", "ArrayPosition", "ArrayJoin", "ArraysOverlap", "ArrayContains",
    "MapFromArraysColumn", "MapColumn", "ArrayColumn", "Slice", "Sequence"
]



from .dataframe import DataFrame
from .internal_utils.readers import InternalReader
from .internal_utils.readwrite import OptionUtils
from .internal_utils.writers import CSVWriter, InternalWriter, JSONWriter
from .utils import IllegalArgumentException

__all__ = ['DataFrameReader', 'DataFrameWriter']


WRITE_MODES = ("overwrite", "append", "ignore", "error", "errorifexists")
DATA_WRITERS = dict(
    csv=CSVWriter,
    json=JSONWriter,
)


class DataFrameReader(OptionUtils):
    def option(self, key, value):
        self._jreader.option(key, value)
        return self

    def schema(self, schema):
        self._jreader.schema(schema)
        return self

    def __init__(self, spark):
        """

        :type spark: gelanis.sql.session.SparkSession
        """
        self._jreader = InternalReader(spark)
        self._spark = spark

    def _df(self, jdf):
        return DataFrame(jdf, self._spark)

    def _generic_read(self, path, method_to_read_if_rdd):
        if isinstance(path, str):
            path = [path]

        if isinstance(path, list):
            return self._df(method_to_read_if_rdd(path))

        # pylint: disable=import-outside-toplevel, cyclic-import
        from ..rdd import RDD

        if isinstance(path, RDD):
            return self._df(method_to_read_if_rdd(path.collect()))

        raise TypeError("path can be only string, list or RDD")

    # pylint: disable=R0914
    def csv(self, path, schema=None, sep=None, encoding=None, quote=None, escape=None,
            comment=None, header=None, inferSchema=None, ignoreLeadingWhiteSpace=None,
            ignoreTrailingWhiteSpace=None, nullValue=None, nanValue=None, positiveInf=None,
            negativeInf=None, dateFormat=None, timestampFormat=None, maxColumns=None,
            maxCharsPerColumn=None, maxMalformedLogPerPartition=None, mode=None,
            columnNameOfCorruptRecord=None, multiLine=None, charToEscapeQuoteEscaping=None,
            samplingRatio=None, enforceSchema=None, emptyValue=None, locale=None, lineSep=None):
        self._set_opts(
            schema=schema, sep=sep, encoding=encoding, quote=quote, escape=escape, comment=comment,
            header=header, inferSchema=inferSchema, ignoreLeadingWhiteSpace=ignoreLeadingWhiteSpace,
            ignoreTrailingWhiteSpace=ignoreTrailingWhiteSpace, nullValue=nullValue,
            nanValue=nanValue, positiveInf=positiveInf, negativeInf=negativeInf,
            dateFormat=dateFormat, timestampFormat=timestampFormat, maxColumns=maxColumns,
            maxCharsPerColumn=maxCharsPerColumn,
            maxMalformedLogPerPartition=maxMalformedLogPerPartition, mode=mode,
            columnNameOfCorruptRecord=columnNameOfCorruptRecord, multiLine=multiLine,
            charToEscapeQuoteEscaping=charToEscapeQuoteEscaping, samplingRatio=samplingRatio,
            enforceSchema=enforceSchema, emptyValue=emptyValue, locale=locale, lineSep=lineSep
        )

        return self._generic_read(path, self._jreader.csv)

    # pylint: disable=R0914
    def json(self, path, schema=None, primitivesAsString=None, prefersDecimal=None,
             allowComments=None, allowUnquotedFieldNames=None, allowSingleQuotes=None,
             allowNumericLeadingZero=None, allowBackslashEscapingAnyCharacter=None,
             mode=None, columnNameOfCorruptRecord=None, dateFormat=None, timestampFormat=None,
             multiLine=None, allowUnquotedControlChars=None, lineSep=None, samplingRatio=None,
             dropFieldIfAllNull=None, encoding=None, locale=None):
        self._set_opts(
            schema=schema, primitivesAsString=primitivesAsString, prefersDecimal=prefersDecimal,
            allowComments=allowComments, allowUnquotedFieldNames=allowUnquotedFieldNames,
            allowSingleQuotes=allowSingleQuotes, allowNumericLeadingZero=allowNumericLeadingZero,
            allowBackslashEscapingAnyCharacter=allowBackslashEscapingAnyCharacter,
            mode=mode, columnNameOfCorruptRecord=columnNameOfCorruptRecord, dateFormat=dateFormat,
            timestampFormat=timestampFormat, multiLine=multiLine,
            allowUnquotedControlChars=allowUnquotedControlChars, lineSep=lineSep,
            samplingRatio=samplingRatio, dropFieldIfAllNull=dropFieldIfAllNull, encoding=encoding,
            locale=locale)

        return self._generic_read(path, self._jreader.json)

    def text(self, path, wholetext=False, lineSep=None,
             pathGlobFilter=None, recursiveFileLookup=None):
        self._set_opts(wholetext=wholetext,
                       lineSep=lineSep,
                       pathGlobFilter=pathGlobFilter,
                       recursiveFileLookup=recursiveFileLookup)

        return self._generic_read(path, self._jreader.text)


class DataFrameWriter(OptionUtils):
    def __init__(self, df):
        self._df = df
        self._spark = df.sql_ctx
        self._jwrite = InternalWriter(self._df)

    def mode(self, saveMode):
        if saveMode is None:
            return self
        if saveMode not in WRITE_MODES:
            accepted_modes = "', '".join(WRITE_MODES)
            raise IllegalArgumentException(
                f"Unknown save mode: {saveMode}. Accepted save modes are {accepted_modes}."
            )
        self._jwrite = self._jwrite.mode(saveMode)
        return self

    def format(self, source):
        self._jwrite = self._jwrite.format(source)
        return self

    def option(self, key, value):
        self._jwrite.option(key, value)
        return self

    def options(self, **options):
        for k in options:
            self._jwrite.option(k, options[k])
        return self

    def partitionBy(self, *cols):
        if len(cols) == 1 and isinstance(cols[0], (list, tuple)):
            cols = cols[0]
        self._jwrite = self._jwrite.partitionBy(cols)
        return self

    def bucketBy(self, numBuckets, col, *cols):
        if not isinstance(numBuckets, int):
            raise TypeError(f"numBuckets should be an int, got {type(numBuckets)}.")

        if isinstance(col, (list, tuple)):
            if cols:
                raise ValueError(f"col is a {type(col)} but cols are not empty")

            cols = col
        else:
            cols = [col] + list(cols)

        if not all(isinstance(c, str) for c in cols):
            raise TypeError("all names should be `str`")

        self._jwrite = self._jwrite.bucketBy(numBuckets, *cols)
        return self

    def sortBy(self, col, *cols):
        if isinstance(col, (list, tuple)):
            if cols:
                raise ValueError(f"col is a {type(col)} but cols are not empty")

            col, cols = col[0], col[1:]

        if not all(isinstance(c, str) for c in cols) or not isinstance(col, str):
            raise TypeError("all names should be `str`")

        self._jwrite = self._jwrite.sortBy(col, *cols)
        return self

    # noinspection PyShadowingBuiltins
    # pylint: disable=W0622
    def save(self, path=None, format=None, mode=None, partitionBy=None, **options):
        self.mode(mode).options(**options)
        if partitionBy is not None:
            self.partitionBy(partitionBy)
        if format is not None:
            self.format(format)
        writer_class = DATA_WRITERS[self._jwrite._source]
        if path is None:
            self._jwrite.save(writer_class)
        else:
            self._jwrite.save(writer_class, path)

    def json(self, path, mode=None, compression=None, dateFormat=None, timestampFormat=None,
             lineSep=None, encoding=None):
        self.mode(mode)
        self._set_opts(
            compression=compression, dateFormat=dateFormat, timestampFormat=timestampFormat,
            lineSep=lineSep, encoding=encoding)
        self.format("json").save(path)

    # pylint: disable=R0914
    def csv(self, path, mode=None, compression=None, sep=None, quote=None, escape=None,
            header=None, nullValue=None, escapeQuotes=None, quoteAll=None, dateFormat=None,
            timestampFormat=None, ignoreLeadingWhiteSpace=None, ignoreTrailingWhiteSpace=None,
            charToEscapeQuoteEscaping=None, encoding=None, emptyValue=None, lineSep=None):
        self.mode(mode)
        self._set_opts(compression=compression, sep=sep, quote=quote, escape=escape, header=header,
                       nullValue=nullValue, escapeQuotes=escapeQuotes, quoteAll=quoteAll,
                       dateFormat=dateFormat, timestampFormat=timestampFormat,
                       ignoreLeadingWhiteSpace=ignoreLeadingWhiteSpace,
                       ignoreTrailingWhiteSpace=ignoreTrailingWhiteSpace,
                       charToEscapeQuoteEscaping=charToEscapeQuoteEscaping,
                       encoding=encoding, emptyValue=emptyValue, lineSep=lineSep)
        self.format("csv").save(path)

    def insertInto(self, tableName=None):
        raise NotImplementedError("Pysparkling does not implement write to table yet")

    def saveAsTable(self, name):
        raise NotImplementedError("Pysparkling does not implement write to table yet")

    def parquet(self, path):
        raise NotImplementedError("Pysparkling does not implement write to parquet yet")

    def text(self, path):
        raise NotImplementedError("Pysparkling does not implement write to text yet")

    def orc(self, path):
        raise NotImplementedError("Pysparkling does not implement write to ORC")

    def jdbc(self, path):
        raise NotImplementedError("Pysparkling does not implement write to JDBC")



from .session import SparkSession


class SQLContext:
    _instantiatedContext = None

    def __init__(self, sparkContext, sparkSession=None, jsqlContext=None):
        self._sc = sparkContext
        if sparkSession is None:
            sparkSession = SparkSession.builder.getOrCreate()
        self.sparkSession = sparkSession
        if SQLContext._instantiatedContext is None:
            SQLContext._instantiatedContext = self

    @classmethod
    def getOrCreate(cls, sc):
        """
        Get the existing SQLContext or create a new one with given Context.

        :param sc: Context
        """
        if cls._instantiatedContext is None:
            cls(sc, SparkSession(sc), None)
        return cls._instantiatedContext

    def newSession(self):
        """
        Returns a new SQLContext as new session, that has separate SQLConf,
        registered temporary views and UDFs, but shared Context and
        table cache.
        """
        return self.__class__(self._sc, self.sparkSession.newSession())

    def setConf(self, key, value):
        """Sets the given Spark SQL configuration property.
        """
        self.sparkSession.conf.set(key, value)



from .. import Expression
from ...types import ArrayType


class Aggregation(Expression):
    @property
    def is_an_aggregation(self):
        return True

    def merge(self, row, schema):
        raise NotImplementedError

    def mergeStats(self, other, schema):
        raise NotImplementedError

    def eval(self, row, schema):
        raise NotImplementedError

    def args(self):
        raise NotImplementedError

    def data_type(self, schema):
        # TODO: Check if we can generalize this. By default, this should be fine, but needs to be overridden in each
        #    subclass where it deviates from this standard.
        # pylint: disable=E1101
        return ArrayType(
            elementType=schema[str(self.column)].dataType,
            containsNull=self.column.is_nullable
        )



import datetime

from dateutil.relativedelta import relativedelta
import pytz

from ...utils import parse_tz
from .._casts import get_time_formatter
from .._date_time import get_unix_timestamp_parser
from ..types import DateType, FloatType, TimestampType
from .expressions import Expression, UnaryExpression
from .operators import Cast

GMT_TIMEZONE = pytz.timezone("GMT")

DAYS_OF_WEEK = ("MON", "TUE", "WED", "THU", "FRI", "SAT", "SUN")


class AddMonths(Expression):
    pretty_name = "add_months"

    def __init__(self, start_date, num_months):
        super().__init__(start_date)
        self.start_date = start_date
        self.num_months = num_months.get_literal_value()
        self.timedelta = datetime.timedelta(days=self.num_months)

    def eval(self, row, schema):
        return Cast(self.start_date, DateType()).eval(row, schema) + self.timedelta

    def args(self):
        return (
            self.start_date,
            self.num_months
        )


class DateAdd(Expression):
    pretty_name = "date_add"

    def __init__(self, start_date, num_days):
        super().__init__(start_date)
        self.start_date = start_date
        self.num_days = num_days.get_literal_value()
        self.timedelta = datetime.timedelta(days=self.num_days)

    def eval(self, row, schema):
        return Cast(self.start_date, DateType()).eval(row, schema) + self.timedelta

    def args(self):
        return (
            self.start_date,
            self.num_days
        )


class DateSub(Expression):
    pretty_name = "date_sub"

    def __init__(self, start_date, num_days):
        super().__init__(start_date)
        self.start_date = start_date
        self.num_days = num_days.get_literal_value()
        self.timedelta = datetime.timedelta(days=self.num_days)

    def eval(self, row, schema):
        return Cast(self.start_date, DateType()).eval(row, schema) - self.timedelta

    def args(self):
        return (
            self.start_date,
            self.num_days
        )


class Year(UnaryExpression):
    pretty_name = "year"

    def eval(self, row, schema):
        return Cast(self.column, DateType()).eval(row, schema).year


class Month(UnaryExpression):
    pretty_name = "month"

    def eval(self, row, schema):
        return Cast(self.column, DateType()).eval(row, schema).month


class Quarter(UnaryExpression):
    pretty_name = "quarter"

    def eval(self, row, schema):
        month = Cast(self.column, DateType()).eval(row, schema).month
        return 1 + int((month - 1) / 3)


class Hour(UnaryExpression):
    pretty_name = "hour"

    def eval(self, row, schema):
        return Cast(self.column, TimestampType()).eval(row, schema).hour


class Minute(UnaryExpression):
    pretty_name = "minute"

    def eval(self, row, schema):
        return Cast(self.column, TimestampType()).eval(row, schema).minute


class Second(UnaryExpression):
    pretty_name = "second"

    def eval(self, row, schema):
        return Cast(self.column, TimestampType()).eval(row, schema).second


class DayOfMonth(UnaryExpression):
    pretty_name = "dayofmonth"

    def eval(self, row, schema):
        return Cast(self.column, DateType()).eval(row, schema).day


class DayOfYear(UnaryExpression):
    pretty_name = "dayofyear"

    def eval(self, row, schema):
        value = Cast(self.column, DateType()).eval(row, schema)
        day_from_the_first = value - datetime.date(value.year, 1, 1)
        return 1 + day_from_the_first.days


class LastDay(UnaryExpression):
    pretty_name = "last_day"

    def eval(self, row, schema):
        value = Cast(self.column, DateType()).eval(row, schema)
        first_of_next_month = value + relativedelta(months=1, day=1)
        return first_of_next_month - datetime.timedelta(days=1)


class WeekOfYear(UnaryExpression):
    pretty_name = "weekofyear"

    def eval(self, row, schema):
        return Cast(self.column, DateType()).eval(row, schema).isocalendar()[1]


class DayOfWeek(UnaryExpression):
    pretty_name = "dayofweek"

    def eval(self, row, schema):
        date = Cast(self.column, DateType()).eval(row, schema)
        return date.isoweekday() + 1 if date.isoweekday() != 7 else 1


class NextDay(Expression):
    pretty_name = "next_day"

    def __init__(self, column, day_of_week):
        super().__init__(column)
        self.column = column
        self.day_of_week = day_of_week.get_literal_value()

    def eval(self, row, schema):
        value = Cast(self.column, DateType()).eval(row, schema)

        if self.day_of_week.upper() not in DAYS_OF_WEEK:
            return None

        today = value.isoweekday()
        target = DAYS_OF_WEEK.index(self.day_of_week.upper()) + 1
        delta = target - today
        if delta <= 0:
            delta += 7
        return value + datetime.timedelta(days=delta)

    def args(self):
        return (
            self.column,
            self.day_of_week
        )


class MonthsBetween(Expression):
    pretty_name = "months_between"

    def __init__(self, column1, column2, round_off):
        super().__init__(column1, column2)
        self.column1 = column1
        self.column2 = column2
        self.round_off = round_off.get_literal_value()

    def eval(self, row, schema):
        value_1 = Cast(self.column1, TimestampType()).eval(row, schema)
        value_2 = Cast(self.column2, TimestampType()).eval(row, schema)

        if (not isinstance(value_1, datetime.datetime)
                or not isinstance(value_2, datetime.datetime)):
            return None

        one_day = datetime.timedelta(days=1)
        value_1_is_the_last_of_its_month = (value_1.month != (value_1 + one_day).month)
        value_2_is_the_last_of_its_month = (value_2.month != (value_2 + one_day).month)
        if value_1.day == value_2.day or (
                value_1_is_the_last_of_its_month
                and value_2_is_the_last_of_its_month
        ):
            # Special cases where time of day is not consider
            diff = ((value_1.year - value_2.year) * 12
                    + (value_1.month - value_2.month))
        else:
            day_offset = (value_1.day - value_2.day
                          + (value_1.hour - value_2.hour) / 24
                          + (value_1.minute - value_2.minute) / 1440
                          + (value_1.second - value_2.second) / 86400)
            diff = ((value_1.year - value_2.year) * 12
                    + (value_1.month - value_2.month) * 1
                    + day_offset / 31)
        if self.round_off:
            return float(round(diff, 8))
        return float(diff)

    def args(self):
        return (
            self.column1,
            self.column2,
            str(self.round_off).lower()
        )


class DateDiff(Expression):
    pretty_name = "datediff"

    def __init__(self, column1, column2):
        super().__init__(column1, column2)
        self.column1 = column1
        self.column2 = column2

    def eval(self, row, schema):
        value_1 = Cast(self.column1, DateType()).eval(row, schema)
        value_2 = Cast(self.column2, DateType()).eval(row, schema)

        if (not isinstance(value_1, datetime.date)
                or not isinstance(value_2, datetime.date)):
            return None

        return (value_1 - value_2).days

    def args(self):
        return (
            self.column1,
            self.column2
        )


class FromUnixTime(Expression):
    pretty_name = "from_unixtime"

    def __init__(self, column, f):
        super().__init__(column)
        self.column = column
        self.format = f.get_literal_value()
        self.formatter = get_time_formatter(self.format)

    def eval(self, row, schema):
        timestamp = Cast(self.column, FloatType()).eval(row, schema)
        return self.formatter(datetime.datetime.fromtimestamp(timestamp))

    def args(self):
        return (
            self.column,
            self.format
        )


class DateFormat(Expression):
    pretty_name = "date_format"

    def __init__(self, column, f):
        super().__init__(column)
        self.column = column
        self.format = f.get_literal_value()
        self.formatter = get_time_formatter(self.format)

    def eval(self, row, schema):
        timestamp = Cast(self.column, TimestampType()).eval(row, schema)
        return self.formatter(timestamp)

    def args(self):
        return (
            self.column,
            self.format
        )


class CurrentTimestamp(Expression):
    pretty_name = "current_timestamp"

    def __init__(self):
        super().__init__()
        self.current_timestamp = None

    def eval(self, row, schema):
        return self.current_timestamp

    def initialize(self, partition_index):
        super().initialize(partition_index)
        self.current_timestamp = datetime.datetime.now()

    def args(self):
        return ()


class CurrentDate(Expression):
    pretty_name = "current_date"

    def __init__(self):
        super().__init__()
        self.current_timestamp = None

    def eval(self, row, schema):
        return self.current_timestamp.date()

    def initialize(self, partition_index):
        super().initialize(partition_index)
        self.current_timestamp = datetime.datetime.now()

    def args(self):
        return ()


class UnixTimestamp(Expression):
    pretty_name = "unix_timestamp"

    def __init__(self, column, f):
        super().__init__(column)
        self.column = column
        self.format = f.get_literal_value()
        self.parser = get_unix_timestamp_parser(self.format)

    def eval(self, row, schema):
        datetime_as_string = self.column.eval(row, schema)
        return self.parser(datetime_as_string)

    def args(self):
        return (
            self.column,
            self.format
        )


class ParseToTimestamp(Expression):
    pretty_name = "to_timestamp"

    def __init__(self, column, f):
        super().__init__(column)
        self.column = column
        self.format = f.get_literal_value()
        self.parser = get_unix_timestamp_parser(self.format)

    def eval(self, row, schema):
        datetime_as_string = self.column.eval(row, schema)
        return datetime.datetime.fromtimestamp(self.parser(datetime_as_string))

    def args(self):
        if self.format is None:
            return (
                f"'{self.column}'",
            )
        return (
            f"'{self.column}'",
            f"'{self.format}'"
        )


class ParseToDate(Expression):
    pretty_name = "to_date"

    def __init__(self, column, f):
        super().__init__(column)
        self.column = column
        self.format = f.get_literal_value()
        self.parser = get_unix_timestamp_parser(self.format)

    def eval(self, row, schema):
        datetime_as_string = self.column.eval(row, schema)
        return datetime.date.fromtimestamp(self.parser(datetime_as_string))

    def args(self):
        if self.format is None:
            return (
                f"'{self.column}'",
            )
        return (
            f"'{self.column}'",
            f"'{self.format}'"
        )


class TruncDate(Expression):
    pretty_name = "trunc"

    def __init__(self, column, level):
        super().__init__(column)
        self.column = column
        self.level = level.get_literal_value()

    def eval(self, row, schema):
        value = Cast(self.column, DateType()).eval(row, schema)
        if self.level in ('year', 'yyyy', 'yy'):
            return datetime.date(value.year, 1, 1)
        if self.level in ('month', 'mon', 'mm'):
            return datetime.date(value.year, value.month, 1)
        return None

    def args(self):
        return (
            self.column,
            self.level
        )


class TruncTimestamp(Expression):
    pretty_name = "date_trunc"

    def __init__(self, level, column):
        super().__init__(column)
        self.level = level.get_literal_value()
        self.column = column

    def eval(self, row, schema):
        value = Cast(self.column, TimestampType()).eval(row, schema)

        day_truncation = self.truncate_to_day(value)
        if day_truncation:
            return day_truncation

        time_truncated = self.truncate_to_time(value)
        if time_truncated:
            return time_truncated

        return None

    def truncate_to_day(self, value):
        if self.level in ('year', 'yyyy', 'yy'):
            return datetime.datetime(value.year, 1, 1)
        if self.level in ('month', 'mon', 'mm'):
            return datetime.datetime(value.year, value.month, 1)
        if self.level in ('day', 'dd'):
            return datetime.datetime(value.year, value.month, value.day)
        if self.level in ('quarter',):
            quarter_start_month = int((value.month - 1) / 3) * 3 + 1
            return datetime.datetime(value.year, quarter_start_month, 1)
        if self.level in ('week',):
            return datetime.datetime(
                value.year, value.month, value.day
            ) - datetime.timedelta(days=value.isoweekday() - 1)
        return None

    def truncate_to_time(self, value):
        if self.level in ('hour',):
            return datetime.datetime(value.year, value.month, value.day, value.hour)
        if self.level in ('minute',):
            return datetime.datetime(value.year, value.month, value.day, value.hour, value.minute)
        if self.level in ('second',):
            return datetime.datetime(
                value.year, value.month, value.day, value.hour, value.minute, value.second
            )
        return None

    def args(self):
        return (
            self.level,
            self.column
        )


class FromUTCTimestamp(Expression):
    pretty_name = "from_utc_timestamp"

    def __init__(self, column, tz):
        super().__init__(column)
        self.column = column
        self.tz = tz.get_literal_value()
        self.pytz = parse_tz(self.tz)

    def eval(self, row, schema):
        value = Cast(self.column, TimestampType()).eval(row, schema)
        if self.pytz is None:
            return value
        gmt_date = GMT_TIMEZONE.localize(value)
        local_date = gmt_date.astimezone(self.pytz)
        return local_date.replace(tzinfo=None)

    def args(self):
        return (
            self.column,
            self.tz
        )


class ToUTCTimestamp(Expression):
    pretty_name = "to_utc_timestamp"

    def __init__(self, column, tz):
        super().__init__(column)
        self.column = column
        self.tz = tz.get_literal_value()
        self.pytz = parse_tz(self.tz)

    def eval(self, row, schema):
        value = Cast(self.column, TimestampType()).eval(row, schema)
        if self.pytz is None:
            return value
        local_date = self.pytz.localize(value)
        gmt_date = local_date.astimezone(GMT_TIMEZONE)
        return gmt_date.replace(tzinfo=None)

    def args(self):
        return (
            self.column,
            self.tz
        )


__all__ = [
    "ToUTCTimestamp", "FromUTCTimestamp", "TruncTimestamp", "TruncDate", "ParseToDate",
    "ParseToTimestamp", "UnixTimestamp", "CurrentTimestamp", "FromUnixTime", "WeekOfYear",
    "NextDay", "MonthsBetween", "LastDay", "DayOfYear", "DayOfMonth", "DayOfWeek", "Month",
    "Quarter", "Year", "DateDiff", "DateSub", "DateAdd", "DateFormat", "CurrentDate",
    "AddMonths", "Hour", "Minute", "Second"
]



"""
The following constants are used to identify join types
"""
INNER_JOIN = "inner"
CROSS_JOIN = "cross"
FULL_JOIN = "full"
LEFT_JOIN = "left"
RIGHT_JOIN = "right"
LEFT_SEMI_JOIN = "leftsemi"
LEFT_ANTI_JOIN = "leftanti"

JOIN_TYPES = dict(
    inner=INNER_JOIN,
    cross=CROSS_JOIN,
    outer=FULL_JOIN,
    full=FULL_JOIN,
    fullouter=FULL_JOIN,
    left=LEFT_JOIN,
    leftouter=LEFT_JOIN,
    right=RIGHT_JOIN,
    rightouter=RIGHT_JOIN,
    leftsemi=LEFT_SEMI_JOIN,
    leftanti=LEFT_ANTI_JOIN,
)



import itertools
import re

from ._row import create_row, Row, row_from_keyed_values
from .internal_utils.joins import (
    CROSS_JOIN, FULL_JOIN, INNER_JOIN, LEFT_ANTI_JOIN, LEFT_JOIN, LEFT_SEMI_JOIN, RIGHT_JOIN
)
from .utils import IllegalArgumentException


def merge_rows_joined_on_values(left, right, left_schema, right_schema, how, on):
    left_names = left_schema.names
    right_names = right_schema.names

    left_on_fields, right_on_fields = get_on_fields(left_schema, right_schema, on)

    on_parts = [
        (on_field, left[on_field] if left is not None else right[on_field])
        for on_field in on
    ]

    if left is None and how in (FULL_JOIN, RIGHT_JOIN):
        left = create_row(left_names, [None for _ in left_names])
    if right is None and how in (LEFT_JOIN, FULL_JOIN):
        right = create_row(right_names, [None for _ in right_names])

    left_parts = (
        (field.name, value)
        for field, value in zip(left_schema.fields, left)
        if field not in left_on_fields
    )

    if how in (INNER_JOIN, CROSS_JOIN, LEFT_JOIN, FULL_JOIN, RIGHT_JOIN):
        right_parts = (
            (field.name, value)
            for field, value in zip(right_schema.fields, right)
            if field not in right_on_fields
        )
    elif how in (LEFT_SEMI_JOIN, LEFT_ANTI_JOIN):
        right_parts = ()
    else:
        raise IllegalArgumentException(f"Argument 'how' cannot be '{how}'")

    return row_from_keyed_values(itertools.chain(on_parts, left_parts, right_parts))


def merge_rows(left, right):
    return create_row(
        itertools.chain(left.__fields__, right.__fields__),
        left + right
    )


FULL_WIDTH_REGEX = re.compile(
    "["
    + r"\u1100-\u115F"
    + r"\u2E80-\uA4CF"
    + r"\uAC00-\uD7A3"
    + r"\uF900-\uFAFF"
    + r"\uFE10-\uFE19"
    + r"\uFE30-\uFE6F"
    + r"\uFF00-\uFF60"
    + r"\uFFE0-\uFFE6"
    + "]"
)


def str_half_width(value):
    """
    Compute string length with full width characters counting for 2 normal ones
    """
    string = format_cell(value)
    if string is None:
        return 0
    if not isinstance(string, str):
        string = str(string)
    return len(string) + len(FULL_WIDTH_REGEX.findall(string))


def pad_cell(cell, truncate, col_width):
    """
    Compute how to pad the value "cell" truncated to truncate so that it fits col width
    :param cell: Any
    :param truncate: int
    :param col_width: int
    :return:
    """
    cell = format_cell(cell)
    cell_width = col_width - str_half_width(cell) + len(cell)
    if truncate > 0:
        return cell.rjust(cell_width)
    return cell.ljust(cell_width)


def format_cell(value):
    """
    Convert a cell value to a string using the logic needed in DataFrame.show()
    """
    if value is None:
        return "null"
    if isinstance(value, bool):
        return str(value).lower()
    if isinstance(value, Row):
        return f"[{', '.join(format_cell(sub_value) for sub_value in value)}]"
    if isinstance(value, dict):
        return "[{0}]".format(
            ", ".join(
                f"{format_cell(key)} -> {format_cell(sub_value)}"
                for key, sub_value in value.items()
            )
        )
    return str(value)


def get_on_fields(left_schema, right_schema, on):
    left_on_fields = [next(field for field in left_schema if field.name == c) for c in on]
    right_on_fields = [next(field for field in right_schema if field.name == c) for c in on]
    return left_on_fields, right_on_fields



from io import BytesIO
import logging

from . import codec, fs

log = logging.getLogger(__name__)


class File:
    """File object.

    :param file_name: Any file name.
    """

    def __init__(self, file_name):
        self.file_name = file_name
        self.fs = fs.get_fs(file_name)(file_name)
        self.codec = codec.get_codec(file_name)()

    @staticmethod
    def resolve_filenames(all_expr):
        """resolve expression for a filename

        :param all_expr:
            A comma separated list of expressions. The expressions can contain
            the wildcard characters ``*`` and ``?``. It also resolves Spark
            datasets to the paths of the individual partitions
            (i.e. ``my_data`` gets resolved to
            ``[my_data/part-00000, my_data/part-00001]``).

        :returns: A list of file names.
        :rtype: list
        """
        files = []
        for expr in all_expr.split(','):
            expr = expr.strip()
            files += fs.get_fs(expr).resolve_filenames(expr)
        log.debug('Filenames: %s', files)
        return files

    @classmethod
    def get_content(cls, all_expr):
        """Return all files matching or in folder matching one of the given expression

        :param all_expr:
            A list of expressions.
            The expressions can contain the wildcard characters ``*`` and ``?``.

        :returns: A list of file names.
        :rtype: list
        """
        files = []
        for expr in all_expr:
            expr = expr.strip()
            files += fs.get_fs(expr).resolve_content(expr)
        log.debug('Filenames: %s', files)
        return files

    def exists(self):
        """Checks both for a file or directory at this location.

        :returns: True or false.
        """
        return self.fs.exists()

    def load(self):
        """Load the data from a file.

        :rtype: io.BytesIO
        """
        stream = self.fs.load()
        stream = self.codec.decompress(stream)
        return stream

    def dump(self, stream=None):
        """Writes a stream to a file.

        :param stream:
            A BytesIO instance. ``bytes`` are also possible and are converted
            to BytesIO.

        :rtype: File
        """
        if stream is None:
            stream = BytesIO()

        if isinstance(stream, bytes):
            stream = BytesIO(stream)

        stream = self.codec.compress(stream)
        self.fs.dump(stream)

        return self

    def make_public(self, recursive=False):
        """Makes the file public. Currently only supported on S3.

        :param recursive: Whether to apply this recursively.
        :rtype: File
        """
        self.fs.make_public(recursive)
        return self



import ast
from functools import partial
import logging

from sqlparser import string_to_ast

from ...sql import functions
from .._expressions.expressions import expression_registry
from .._expressions.literals import Literal
from .._expressions.mappers import Concat, CreateStruct
from .._expressions.operators import (
    Add, Alias, And, BitwiseAnd, BitwiseNot, BitwiseOr, BitwiseXor, Cast, Divide, Equal, GreaterThan,
    GreaterThanOrEqual, Invert, LessThan, LessThanOrEqual, Minus, Mod, Negate, Or, Time, UnaryPositive
)
from ..column import Column, parse
from ..types import DoubleType, parsed_string_to_type, StringType, StructField, StructType
from ..utils import ParseException


class SqlParsingError(ParseException):
    pass


class UnsupportedStatement(SqlParsingError):
    pass


def check_children(expected, children):
    if len(children) != expected:
        raise SqlParsingError(
            "Expecting {0} children, got {1}: {2}".format(
                expected, len(children), children
            )
        )


def unwrap(*children, has_quotes=False):
    check_children(1, children)
    tmp = convert_tree(children[0])

    if has_quotes:
        return tmp[1:-1]  # Strip the quotes.

    return tmp


def never_found(*children):
    logging.warning("We should never have encounter this node.")
    return unwrap(*children)


def unsupported(*children):
    raise UnsupportedStatement


def add_comment(*children):
    # children[0] == 'COMMENT'
    return {'comment': get_leaf_value(children[1])[1:-1]}  # [1:-1] to strip the quotes.


def empty(*children):
    check_children(0, children)


def first_child_only(*children):
    return convert_tree(children[0])


def child_and_eof(*children):
    check_children(2, children)
    return convert_tree(children[0])


def convert_tree(tree):
    tree_type = tree.__class__.__name__
    if not hasattr(tree, "children"):
        return get_leaf_value(tree)
    try:
        converter = CONVERTERS[tree_type]
    except UnsupportedStatement:
        raise SqlParsingError("Unsupported statement {0}".format(tree_type)) from None

    if tree.children is None:
        raise SqlParsingError

    return converter(*tree.children)


def call_function(*children):
    raw_function_name = convert_tree(children[0])
    function_name = next(
        (name for name in functions.__all__ if name.lower() == raw_function_name.lower()),
        None
    )
    function_expression = expression_registry.get(function_name.lower())
    params = [convert_tree(c) for c in children[2:-1]]

    # todo: there must be a cleaner way to do that
    complex_function = any(
        not isinstance(param, Column) and param == ')'
        for param in params
    )
    if not complex_function:
        last_argument_position = None
        # filter_clause = None
        # over_clause = None
        # set_clause = None
    else:
        last_argument_position = params.index(")")
        # filter_clause = ...  # todo
        # over_clause = ...  # todo
        # set_clause = ...  # todo

    # parameters are comma separated
    function_arguments = params[0:last_argument_position:2]
    return function_expression(*function_arguments)


def binary_operation(*children):
    check_children(3, children)
    left, operator, right = children
    cls = binary_operations[convert_tree(operator).upper()]
    return cls(
        convert_tree(left),
        convert_tree(right)
    )


def cast_context(*children):
    """
    Children are:
    CAST '(' expression AS dataType ')'

    """
    check_children(6, children)
    expression = convert_tree(children[2])
    data_type = convert_tree(children[4])
    return parse(expression).cast(data_type)


def detect_data_type(*children):
    if children[0].__class__.__name__ == 'ErrorNodeImpl':
        children = children[1:]

    if any(child.__class__.__name__ == 'ErrorNodeImpl' for child in children):
        raise SqlParsingError

    data_type = convert_tree(children[0])
    params = [convert_tree(c) for c in children[2:-1:2]]
    return parsed_string_to_type(data_type, params)


def unary_operation(*children):
    check_children(2, children)
    operator, value = children
    cls = unary_operations[convert_tree(operator).upper()]
    return cls(
        convert_tree(value)
    )


def parenthesis_context(*children):
    check_children(3, children)
    return convert_tree(children[1])


def convert_boolean(*children):
    check_children(1, children)
    value = convert_tree(children[0])
    if value.lower() == "false":
        return False
    if value.lower() == "true":
        return True
    raise SqlParsingError("Expecting boolean value, got {0}".format(children))


def convert_to_literal(*children):
    check_children(1, children)
    value = convert_tree(children[0])
    return Literal(value)


def convert_column(*children):
    check_children(1, children)
    value = convert_tree(children[0])
    return Column(value)


def convert_field(*children):
    name = convert_tree(children[0])
    data_type = convert_tree(children[1])
    return StructField(name, data_type)


def convert_table_schema(*children):
    check_children(2, children)
    return StructType(fields=convert_tree(children[0]))


def convert_to_null(*children):
    return None


def get_leaf_value(*children):
    check_children(1, children)
    value = children[0]

    if value.__class__.__name__ == 'NonReservedContext':
        return get_leaf_value(*value.children)

    if value.__class__.__name__ != "TerminalNodeImpl":
        raise SqlParsingError("Expecting TerminalNodeImpl, got {0}".format(type(value).__name__))

    if not hasattr(value, "symbol"):
        raise SqlParsingError("Got leaf value but without symbol")

    return value.symbol.text


def remove_delimiter(*children):
    delimited_value = get_leaf_value(*children)
    return delimited_value[1:-1]


def explicit_list(*children):
    return tuple(
        convert_tree(c)
        for c in children[1:-1:2]
    )


def implicit_list(*children):
    return tuple(
        convert_tree(c)
        for c in children
        if c.__class__.__name__ != 'TerminalNodeImpl'
    )


def concat_to_value(*children):
    return ast.literal_eval("".join(convert_tree(c) for c in children))


def concat_keywords(*children):
    return " ".join(convert_tree(c) for c in children)


def concat_strings(*children):
    return "".join(convert_tree(c) for c in children)


def build_struct(*children):
    return CreateStruct(*(convert_tree(c) for c in children[2:-1:2]))


def potential_alias(*children):
    if len(children) == 1:
        return convert_tree(children[0])
    if len(children) in (2, 3):
        return Alias(
            convert_tree(children[0]),
            convert_tree(children[-1])
        )
    raise SqlParsingError("Expecting 1, 2 or 3 children, got {0}".format(len(children)))


def check_identifier(*children):
    check_children(2, children)
    identifier = convert_tree(children[0])
    if children[1].children:
        extra = convert_tree(children[1])
        raise SqlParsingError(
            "Possibly unquoted identifier {0}{1} detected. "
            "Please consider quoting it with back-quotes as `{0}{1}`".format(identifier, extra)
        )
    return identifier


CONVERTERS = {
    'AddTableColumnsContext': unsupported,
    'AddTablePartitionContext': unsupported,
    'AggregationClauseContext': unsupported,
    'AliasedQueryContext': unsupported,
    'AliasedRelationContext': unsupported,
    'AlterColumnActionContext': unsupported,
    'AlterTableAlterColumnContext': unsupported,
    'AlterViewQueryContext': unsupported,
    'AnalyzeContext': unsupported,
    'AnsiNonReservedContext': get_leaf_value,
    'ApplyTransformContext': unsupported,
    'ArithmeticBinaryContext': binary_operation,
    'ArithmeticOperatorContext': get_leaf_value,
    'ArithmeticUnaryContext': unary_operation,
    'AssignmentContext': unsupported,
    'AssignmentListContext': implicit_list,
    'BigDecimalLiteralContext': concat_to_value,
    'BigIntLiteralContext': concat_to_value,
    'BooleanExpressionContext': never_found,
    'BooleanLiteralContext': unwrap,
    'BooleanValueContext': convert_boolean,
    'BucketSpecContext': unsupported,
    'CacheTableContext': unsupported,
    'CastContext': cast_context,
    'ClearCacheContext': unsupported,
    'ColPositionContext': unsupported,
    'ColTypeContext': convert_field,
    'ColTypeListContext': implicit_list,
    'ColumnReferenceContext': convert_column,
    'CommentNamespaceContext': unsupported,
    'CommentSpecContext': add_comment,
    'CommentTableContext': unsupported,
    'ComparisonContext': binary_operation,
    'ComparisonOperatorContext': get_leaf_value,
    'ComplexColTypeContext': implicit_list,
    'ComplexColTypeListContext': implicit_list,
    'ComplexDataTypeContext': detect_data_type,
    'ConstantContext': never_found,
    'ConstantDefaultContext': convert_to_literal,
    'ConstantListContext': explicit_list,
    'CreateFileFormatContext': unsupported,
    'CreateFunctionContext': unsupported,
    'CreateHiveTableContext': unsupported,
    'CreateNamespaceContext': unsupported,
    'CreateTableClausesContext': unsupported,
    'CreateTableContext': unsupported,
    'CreateTableHeaderContext': unsupported,
    'CreateTableLikeContext': unsupported,
    'CreateTempViewUsingContext': unsupported,
    'CreateViewContext': unsupported,
    'CtesContext': unsupported,
    'CurrentDatetimeContext': unsupported,
    'DataTypeContext': never_found,
    'DecimalLiteralContext': concat_to_value,
    'DeleteFromTableContext': unsupported,
    'DereferenceContext': unsupported,
    'DescribeColNameContext': unsupported,
    'DescribeFuncNameContext': unwrap,
    'DescribeFunctionContext': unsupported,
    'DescribeNamespaceContext': unsupported,
    'DescribeQueryContext': unsupported,
    'DescribeRelationContext': unsupported,
    'DmlStatementContext': unsupported,
    'DmlStatementNoWithContext': never_found,
    'DoubleLiteralContext': concat_to_value,
    'DropFunctionContext': unsupported,
    'DropNamespaceContext': unsupported,
    'DropTableColumnsContext': unsupported,
    'DropTableContext': unsupported,
    'DropTablePartitionsContext': unsupported,
    'DropViewContext': unsupported,
    'ErrorCapturingIdentifierContext': check_identifier,
    'ErrorCapturingIdentifierExtraContext': never_found,
    'ErrorCapturingMultiUnitsIntervalContext': unsupported,
    'ErrorCapturingUnitToUnitIntervalContext': unsupported,
    'ErrorIdentContext': concat_strings,
    'ExistsContext': unsupported,
    'ExplainContext': unsupported,
    'ExponentLiteralContext': concat_to_value,
    'ExpressionContext': unwrap,
    'ExtractContext': unsupported,
    'FailNativeCommandContext': unsupported,
    'FileFormatContext': never_found,
    'FirstContext': unsupported,
    'FrameBoundContext': unsupported,
    'FromClauseContext': unsupported,
    'FromStatementBodyContext': unsupported,
    'FromStatementContext': unsupported,
    'FromStmtContext': unwrap,
    'FunctionCallContext': call_function,
    'FunctionIdentifierContext': unsupported,
    'FunctionNameContext': unwrap,
    'FunctionTableContext': unsupported,
    'GenericFileFormatContext': unwrap,
    'GroupingSetContext': unsupported,
    'HavingClauseContext': unsupported,
    'HintContext': unsupported,
    'HintStatementContext': unsupported,
    'HiveChangeColumnContext': unsupported,
    'HiveReplaceColumnsContext': unsupported,
    'IdentifierCommentContext': unsupported,
    'IdentifierCommentListContext': explicit_list,
    'IdentifierContext': unwrap,
    'IdentifierListContext': explicit_list,
    'IdentifierSeqContext': unsupported,
    'IdentityTransformContext': unwrap,
    'InlineTableContext': unsupported,
    'InlineTableDefault1Context': unwrap,
    'InlineTableDefault2Context': unwrap,
    'InsertIntoContext': never_found,
    'InsertIntoTableContext': unsupported,
    'InsertOverwriteDirContext': unsupported,
    'InsertOverwriteHiveDirContext': unsupported,
    'InsertOverwriteTableContext': unsupported,
    'IntegerLiteralContext': concat_to_value,
    'IntervalContext': unsupported,
    'IntervalLiteralContext': unwrap,
    'IntervalUnitContext': unwrap,
    'IntervalValueContext': unsupported,
    'JoinCriteriaContext': unsupported,
    'JoinRelationContext': unsupported,
    'JoinTypeContext': concat_keywords,
    'LambdaContext': unsupported,
    'LastContext': unsupported,
    'LateralViewContext': unsupported,
    'LegacyDecimalLiteralContext': concat_to_value,
    'LoadDataContext': unsupported,
    'LocationSpecContext': unsupported,
    'LogicalBinaryContext': binary_operation,
    'LogicalNotContext': unary_operation,
    'ManageResourceContext': unsupported,
    'MatchedActionContext': unsupported,
    'MatchedClauseContext': unsupported,
    'MergeIntoTableContext': unsupported,
    'MultiInsertQueryBodyContext': unsupported,
    'MultiInsertQueryContext': unsupported,
    'MultipartIdentifierContext': unsupported,
    'MultipartIdentifierListContext': implicit_list,
    'MultiUnitsIntervalContext': unsupported,
    'NamedExpressionContext': potential_alias,
    'NamedExpressionSeqContext': unsupported,
    'NamedQueryContext': unsupported,
    'NamedWindowContext': unsupported,
    'NamespaceContext': get_leaf_value,
    'NestedConstantListContext': explicit_list,
    'NonReservedContext': get_leaf_value,
    'NotMatchedActionContext': unsupported,
    'NotMatchedClauseContext': unsupported,
    'NullLiteralContext': convert_to_null,
    'NumberContext': never_found,
    'NumericLiteralContext': unwrap,
    'OrderedIdentifierContext': unsupported,
    'OrderedIdentifierListContext': explicit_list,
    'OverlayContext': unsupported,
    'ParenthesizedExpressionContext': parenthesis_context,
    'PartitionSpecContext': unsupported,
    'PartitionSpecLocationContext': unsupported,
    'PartitionValContext': unsupported,
    'PivotClauseContext': unsupported,
    'PivotColumnContext': unsupported,
    'PivotValueContext': unsupported,
    'PositionContext': unsupported,
    'PredicateContext': unsupported,
    'PredicatedContext': unwrap,
    'PredicateOperatorContext': get_leaf_value,
    'PrimaryExpressionContext': never_found,
    'PrimitiveDataTypeContext': detect_data_type,
    'QualifiedColTypeWithPositionContext': unsupported,
    'QualifiedColTypeWithPositionListContext': implicit_list,
    'QualifiedNameContext': concat_strings,
    'QualifiedNameListContext': implicit_list,
    'QueryContext': unsupported,
    'QueryOrganizationContext': unsupported,
    'QueryPrimaryContext': never_found,
    'QueryPrimaryDefaultContext': unwrap,
    'QuerySpecificationContext': unsupported,
    'QueryTermContext': never_found,
    'QueryTermDefaultContext': unwrap,
    'QuotedIdentifierAlternativeContext': partial(unwrap, has_quotes=True),
    'QuotedIdentifierContext': get_leaf_value,
    'RealIdentContext': empty,
    'RecoverPartitionsContext': unsupported,
    'RefreshResourceContext': unsupported,
    'RefreshTableContext': unsupported,
    'RegularQuerySpecificationContext': unsupported,
    'RelationContext': unsupported,
    'RelationPrimaryContext': never_found,
    'RenameTableColumnContext': unsupported,
    'RenameTableContext': unsupported,
    'RenameTablePartitionContext': unsupported,
    'RepairTableContext': unsupported,
    'ReplaceTableContext': unsupported,
    'ReplaceTableHeaderContext': unsupported,
    'ResetConfigurationContext': unwrap,
    'ResourceContext': unsupported,
    'RowConstructorContext': unsupported,
    'RowFormatContext': never_found,
    'RowFormatDelimitedContext': unsupported,
    'RowFormatSerdeContext': unsupported,
    'SampleByBucketContext': unsupported,
    'SampleByBytesContext': unsupported,
    'SampleByPercentileContext': unsupported,
    'SampleByRowsContext': unsupported,
    'SampleContext': unsupported,
    'SampleMethodContext': never_found,
    'SearchedCaseContext': unsupported,
    'SelectClauseContext': unsupported,
    'SetClauseContext': unsupported,
    'SetConfigurationContext': unsupported,
    'SetNamespaceLocationContext': unsupported,
    'SetNamespacePropertiesContext': unsupported,
    'SetOperationContext': unsupported,
    'SetQuantifierContext': get_leaf_value,
    'SetTableLocationContext': unsupported,
    'SetTablePropertiesContext': unsupported,
    'SetTableSerDeContext': unsupported,
    'ShowColumnsContext': unsupported,
    'ShowCreateTableContext': unsupported,
    'ShowCurrentNamespaceContext': unsupported,
    'ShowFunctionsContext': unsupported,
    'ShowNamespacesContext': unsupported,
    'ShowPartitionsContext': unsupported,
    'ShowTableContext': unsupported,
    'ShowTablesContext': unsupported,
    'ShowTblPropertiesContext': unsupported,
    'ShowViewsContext': unsupported,
    'SimpleCaseContext': unsupported,
    'SingleDataTypeContext': child_and_eof,
    'SingleExpressionContext': child_and_eof,
    'SingleFunctionIdentifierContext': child_and_eof,
    'SingleInsertQueryContext': unsupported,
    'SingleMultipartIdentifierContext': child_and_eof,
    'SingleStatementContext': first_child_only,
    'SingleTableIdentifierContext': child_and_eof,
    'SingleTableSchemaContext': convert_table_schema,
    'SkewSpecContext': unsupported,
    'SmallIntLiteralContext': concat_to_value,
    'SortItemContext': unsupported,
    'SqlBaseParser': unsupported,
    'StarContext': unsupported,
    'StatementContext': never_found,
    'StatementDefaultContext': unwrap,
    'StorageHandlerContext': unsupported,
    'StrictIdentifierContext': never_found,
    'StrictNonReservedContext': get_leaf_value,
    'StringLiteralContext': remove_delimiter,
    'StructContext': build_struct,
    'SubqueryContext': parenthesis_context,
    'SubqueryExpressionContext': parenthesis_context,
    'SubscriptContext': unsupported,
    'SubstringContext': unsupported,
    'TableAliasContext': unsupported,
    'TableContext': unsupported,
    'TableFileFormatContext': unsupported,
    'TableIdentifierContext': unsupported,
    'TableNameContext': unsupported,
    'TablePropertyContext': unsupported,
    'TablePropertyKeyContext': unsupported,
    'TablePropertyListContext': explicit_list,
    'TablePropertyValueContext': unwrap,
    'TableProviderContext': unsupported,
    'onContext': unwrap,
    'TerminalNodeImpl': get_leaf_value,
    'TinyIntLiteralContext': concat_to_value,
    'TransformArgumentContext': unwrap,
    'TransformClauseContext': unsupported,
    'TransformContext': never_found,
    'TransformListContext': explicit_list,
    'TransformQuerySpecificationContext': unsupported,
    'TrimContext': unsupported,
    'TruncateTableContext': unsupported,
    'TypeConstructorContext': unsupported,
    'UncacheTableContext': unsupported,
    'UnitToUnitIntervalContext': unsupported,
    'UnquotedIdentifierContext': get_leaf_value,
    'UnsetTablePropertiesContext': unsupported,
    'UnsupportedHiveNativeCommandsContext': unsupported,
    'UpdateTableContext': unsupported,
    'UseContext': unsupported,
    'ValueExpressionContext': never_found,
    'ValueExpressionDefaultContext': unwrap,
    'WhenClauseContext': unsupported,
    'WhereClauseContext': unsupported,
    'WindowClauseContext': unsupported,
    'WindowDefContext': unsupported,
    'WindowFrameContext': unsupported,
    'WindowRefContext': unsupported,
    'WindowSpecContext': never_found,
}

binary_operations = {
    "=": Equal,
    "==": Equal,
    "<>": lambda *args: Invert(Equal(*args)),
    "!=": lambda *args: Invert(Equal(*args)),
    "<": LessThan,
    "<=": LessThanOrEqual,
    "!>": LessThanOrEqual,
    ">": GreaterThan,
    ">=": GreaterThanOrEqual,
    "!<": GreaterThanOrEqual,
    "+": Add,
    "-": Minus,
    '*': Time,
    '/': lambda a, b: Divide(Cast(a, DoubleType), Cast(b, DoubleType)),
    '%': Mod,
    'DIV': lambda a, b: Divide(Cast(a, DoubleType), Cast(b, DoubleType)),
    '&': BitwiseAnd,
    '|': BitwiseOr,
    '||': lambda a, b: Concat([Cast(a, StringType), Cast(b, StringType)]),
    '^': BitwiseXor,
    'AND': And,
    'OR': Or,
}

unary_operations = {
    "+": UnaryPositive,
    "-": Negate,
    "~": BitwiseNot,
    'NOT': Invert
}


def parse_sql(string, rule, debug=False):
    try:
        tree = string_to_ast(string, rule, debug=debug)
    except TypeError as e:
        raise SqlParsingError(str(e)) from e

    return convert_tree(tree)


def parse_data_type(string, debug=False):
    return parse_sql(string, "singleDataType", debug)


def parse_schema(string, debug=False):
    return parse_sql(string, "singleTableSchema", debug)


def parse_expression(string, debug=False):
    return parse_sql(string, "singleExpression", debug)



_sentinel = object()


class RuntimeConfig:
    def __init__(self, jconf=None):
        self._conf = {}

    def set(self, key, value):
        self._conf[key] = value

    def get(self, key, default=_sentinel):
        self._checkType(key, "key")
        if default is _sentinel:
            return self._conf.get(key)
        if default is not None:
            self._checkType(default, "default")
        return self._conf.get(key, default)

    def unset(self, key):
        del self._conf[key]

    def _checkType(self, obj, identifier):
        if not isinstance(obj, str):
            raise TypeError("expected %s '%s' to be a string (was '%s')" %
                            (identifier, obj, type(obj).__name__))

    def isModifiable(self, key):
        raise NotImplementedError("gelanis does not support yet this feature")



from collections import Counter
from copy import deepcopy
from functools import partial
import itertools
import json
import math
import warnings

from ..storagelevel import StorageLevel
from ..utils import compute_weighted_percentiles, get_keyfunc, portable_hash, reservoir_sample_and_size
from ._row import create_row, Row, row_from_keyed_values
from ._schema_utils import get_schema_from_cols, infer_schema_from_rdd, merge_schemas
from ._statcounter import CovarianceCounter, RowStatHelper
from ._utils import format_cell, merge_rows, merge_rows_joined_on_values, pad_cell, str_half_width
from .column import parse
from .functions import array, collect_set, count, lit, map_from_arrays, rand, struct
from .internal_utils.column import resolve_column
from .internal_utils.joins import (
    CROSS_JOIN, FULL_JOIN, INNER_JOIN, LEFT_ANTI_JOIN, LEFT_JOIN, LEFT_SEMI_JOIN, RIGHT_JOIN
)
from .types import DataType, LongType, StringType, StructField, StructType
from .utils import IllegalArgumentException


def _generate_show_layout(char: str, fields):
    if not fields:
        return ''

    txt = char
    txt += char.join(fields)
    txt += char + '\n'

    return txt


class FieldIdGenerator:
    """
    This metaclass adds an unique ID to all instances of its classes.

    This allows to identify that a field was, when created, associated to a DataFrame.

    Such field can be retrieved with the syntax df.name to build an operation.
    The id clarifies if it is still associated to a field and which one
    when the operation is applied.

    While id() allow the same behaviour in most cases, this one:
    - Allows deep copies which are needed for aggregation
    - Support distributed computation, e.g. multiprocessing
    """
    _id = 0

    @classmethod
    def next(cls):
        cls._id += 1
        return cls._id

    @classmethod
    def bind_schema(cls, schema):
        for field in schema.fields:
            if not hasattr(field, "id"):
                field.id = cls.next()
            if isinstance(field, StructType):
                cls.bind_schema(field)
        return schema

    @classmethod
    def unbind_schema(cls, schema):
        for field in schema.fields:
            delattr(field, "id")
            if isinstance(field, StructType):
                cls.unbind_schema(field)
        return schema


class DataFrameInternal:
    def __init__(self, sc, rdd, cols=None, convert_to_row=False, schema=None):
        """
        :type rdd: RDD
        """
        if convert_to_row:
            if cols is None:
                cols = [f"_c{i}" for i in range(200)]
            rdd = rdd.map(partial(create_row, cols))

        self._sc = sc
        self._rdd = rdd
        if schema is None and convert_to_row is False:
            raise NotImplementedError(
                "Schema cannot be None when creating DataFrameInternal from another. "
                "As a user you should not see this error, feel free to report a bug at "
                "https://github.com/svenkreiss/gelanis/issues"
            )
        if schema is not None:
            self._set_schema(schema)
        else:
            self._set_schema(infer_schema_from_rdd(self._rdd))

    def _set_schema(self, schema):
        bound_schema = FieldIdGenerator.bind_schema(deepcopy(schema))
        self.bound_schema = bound_schema

    @property
    def unbound_schema(self):
        schema = deepcopy(self.bound_schema)
        return FieldIdGenerator.unbind_schema(schema)

    def _with_rdd(self, rdd, schema):
        return DataFrameInternal(
            self._sc,
            rdd,
            schema=schema
        )

    def rdd(self):
        return self._rdd

    @staticmethod
    def range(sc, start, end=None, step=1, numPartitions=None):
        if end is None:
            start, end = 0, start

        rdd = sc.parallelize(
            ([i] for i in range(start, end, step)),
            numSlices=numPartitions
        )
        return DataFrameInternal(sc, rdd, ["id"], True)

    def count(self):
        return self._rdd.count()

    def collect(self):
        return self._rdd.collect()

    def toLocalIterator(self):
        return self._rdd.toLocalIterator()

    def limit(self, n):
        jdf = self._sc.parallelize(self._rdd.take(n))
        return self._with_rdd(jdf, self.bound_schema)

    def take(self, n):
        return self._rdd.take(n)

    def foreach(self, f):
        self._rdd.foreach(f)

    def foreachPartition(self, f):
        self._rdd.foreachPartition(f)

    def cache(self):
        return self._with_rdd(self._rdd.cache(), self.bound_schema)

    def persist(self, storageLevel=StorageLevel.MEMORY_ONLY):
        return self._with_rdd(self._rdd.persist(storageLevel), self.bound_schema)

    def unpersist(self, blocking=False):
        return self._with_rdd(self._rdd.unpersist(blocking), self.bound_schema)

    def coalesce(self, numPartitions):
        return self._with_rdd(self._rdd.coalesce(numPartitions), self.bound_schema)

    def distinct(self):
        return self._with_rdd(self._rdd.distinct(), self.bound_schema)

    def sample(self, withReplacement=None, fraction=None, seed=None):
        return self._with_rdd(
            self._rdd.sample(
                withReplacement=withReplacement,
                fraction=fraction,
                seed=seed
            ),
            self.bound_schema
        )

    def randomSplit(self, weights, seed):
        return self._with_rdd(
            self._rdd.randomSplit(weights=weights, seed=seed),
            self.bound_schema
        )

    @property
    def storageLevel(self):
        return getattr(self._rdd, "storageLevel", StorageLevel(False, False, False, False))

    def is_cached(self):
        return hasattr(self._rdd, "storageLevel")

    def simple_repartition(self, numPartitions):
        return self._with_rdd(self._rdd.repartition(numPartitions), self.bound_schema)

    def repartitionByValues(self, numPartitions, partitioner=None):
        return self._with_rdd(
            self._rdd.map(lambda x: (x, x)).partitionBy(numPartitions, partitioner).values(),
            self.bound_schema
        )

    def repartition(self, numPartitions, cols):
        def partitioner(row):
            return sum(hash(c.eval(row, self.bound_schema)) for c in cols)

        return self.repartitionByValues(numPartitions, partitioner)

    def repartitionByRange(self, numPartitions, *cols):
        key = get_keyfunc(cols, self.bound_schema)
        bounds = self._get_range_bounds(self._rdd, numPartitions, key=key)

        def get_range_id(value):
            return sum(1 for bound in bounds if key(bound) < key(value))

        return self.repartitionByValues(numPartitions, partitioner=get_range_id)

    @staticmethod
    def _get_range_bounds(rdd, numPartitions, key):
        if numPartitions == 0:
            return []

        # pylint: disable=W0511
        # todo: check if sample_size is set in SQLConf.get.rangeExchangeSampleSizePerPartition
        # sample_size = min(
        #   SQLConf.get.rangeExchangeSampleSizePerPartition * rdd.getNumPartitions(),
        #   1e6
        # )
        sample_size = 1e6
        sample_size_per_partition = math.ceil(3 * sample_size / numPartitions)
        sketched_rdd = DataFrameInternal.sketch_rdd(rdd, sample_size_per_partition)
        rdd_size = sum(partition_size for partition_size, sample in sketched_rdd.values())

        if rdd_size == 0:
            return []

        fraction = sample_size / rdd_size

        candidates, imbalanced_partitions = DataFrameInternal._get_initial_candidates(
            sketched_rdd,
            sample_size_per_partition,
            fraction
        )

        additional_candidates = DataFrameInternal._get_additional_candidates(
            rdd,
            imbalanced_partitions,
            fraction
        )

        candidates += additional_candidates
        bounds = compute_weighted_percentiles(
            candidates,
            min(numPartitions, len(candidates)) + 1,
            key=key
        )[1:-1]
        return bounds

    @staticmethod
    def _get_initial_candidates(sketched_rdd, sample_size_per_partition, fraction):
        candidates = []
        imbalanced_partitions = set()
        for idx, (partition_size, sample) in sketched_rdd.items():
            # Partition is bigger than (3 times) average and more than sample_size_per_partition
            # is needed to get accurate information on its distribution
            if fraction * partition_size > sample_size_per_partition:
                imbalanced_partitions.add(idx)
            else:
                # The weight is 1 over the sampling probability.
                weight = partition_size / len(sample)
                candidates += [(key, weight) for key in sample]
        return candidates, imbalanced_partitions

    @staticmethod
    def _get_additional_candidates(rdd, imbalanced_partitions, fraction):
        additional_candidates = []
        if imbalanced_partitions:
            # Re-sample imbalanced partitions with the desired sampling probability.
            def keep_imbalanced_partitions(partition_id, x):
                return x if partition_id in imbalanced_partitions else []

            resampled = (rdd.mapPartitionsWithIndex(keep_imbalanced_partitions)
                         .sample(withReplacement=False, fraction=fraction, seed=rdd.id())
                         .collect())
            weight = (1.0 / fraction).toFloat
            additional_candidates += [(x, weight) for x in resampled]
        return additional_candidates

    @staticmethod
    def sketch_rdd(rdd, sample_size_per_partition):
        """
        Get a subset per partition of an RDD

        Sampling algorithm is reservoir sampling.

        :param rdd:
        :param sample_size_per_partition:
        :return:
        """

        def sketch_partition(idx, x):
            sample, original_size = reservoir_sample_and_size(
                x,
                sample_size_per_partition,
                seed=rdd.id() + idx
            )
            return [(idx, (original_size, sample))]

        sketched_rdd_content = rdd.mapPartitionsWithIndex(sketch_partition).collect()

        return dict(sketched_rdd_content)

    def sampleBy(self, col, fractions, seed):
        fractions_as_col = map_from_arrays(
            array(*(map(lit, fractions.keys()))),
            array(*map(lit, fractions.values()))
        )

        return self._with_rdd(
            self.filter(rand(seed) < fractions_as_col[col]),
            self.bound_schema
        )

    def toJSON(self, use_unicode):
        """

        :rtype: RDD
        """
        return self._rdd.map(lambda row: json.dumps(row.asDict(True), ensure_ascii=not use_unicode))

    def sortWithinPartitions(self, cols, ascending):
        key = get_keyfunc([parse(c) for c in cols], self.bound_schema)

        def partition_sort(data):
            return sorted(data, key=key, reverse=not ascending)

        return self._with_rdd(
            self._rdd.mapPartitions(partition_sort),
            self.bound_schema
        )

    def sort(self, cols):
        # Pysparkling implementation of RDD.sortBy is an in-order sort,
        # calling it multiple times allow sorting
        # based on multiple criteria and ascending orders
        # pylint: disable=W0511
        # Todo: this could be optimized as it's possible to sort
        #  together columns that are in the same ascending order
        sorted_rdd = self._rdd
        for col in cols[::-1]:
            ascending = col.sort_order in ["ASC NULLS FIRST", "ASC NULLS LAST"]
            nulls_are_smaller = col.sort_order in ["DESC NULLS LAST", "ASC NULLS FIRST"]
            key = get_keyfunc([col], self.bound_schema, nulls_are_smaller=nulls_are_smaller)
            sorted_rdd = sorted_rdd.sortBy(key, ascending=ascending)
        return self._with_rdd(sorted_rdd, self.bound_schema)

    def select(self, *exprs):
        cols = [parse(e) for e in exprs]

        if any(col.is_an_aggregation for col in cols):
            df_as_group = InternalGroupedDataFrame(self, [])
            return df_as_group.agg(exprs)

        def select_mapper(partition_index, partition):
            # Initialize non deterministic functions so that they are reproducible
            initialized_cols = [col.initialize(partition_index) for col in cols]
            generators = [col for col in initialized_cols if col.may_output_multiple_rows]
            non_generators = [col for col in initialized_cols if not col.may_output_multiple_rows]
            number_of_generators = len(generators)
            if number_of_generators > 1:
                raise Exception(
                    "Only one generator allowed per select clause"
                    f" but found {number_of_generators}: {', '.join(generators)}"
                )

            return self.get_select_output_field_lists(
                partition,
                non_generators,
                initialized_cols,
                generators[0] if generators else None
            )

        new_schema = get_schema_from_cols(cols, self.bound_schema)
        return self._with_rdd(
            self._rdd.mapPartitionsWithIndex(select_mapper),
            schema=new_schema
        )

    def get_select_output_field_lists(self, partition, non_generators, initialized_cols, generator):
        output_rows = []
        for row in partition:
            base_row_fields = []
            for col in non_generators:
                output_cols, output_values = resolve_column(col, row, schema=self.bound_schema)
                base_row_fields += zip(output_cols, output_values[0])

            if generator is not None:
                generated_row_fields = self.get_generated_row_fields(
                    generator, row, initialized_cols, base_row_fields
                )
                for generated_row in generated_row_fields:
                    output_rows.append(
                        row_from_keyed_values(generated_row, metadata=row.get_metadata())
                    )
            else:
                output_rows.append(
                    row_from_keyed_values(base_row_fields, metadata=row.get_metadata())
                )
        return output_rows

    def get_generated_row_fields(self, generator, row, initialized_cols, base_row):
        additional_fields = []
        generator_position = initialized_cols.index(generator)
        generated_cols, generated_sub_rows = resolve_column(
            generator, row, schema=self.bound_schema
        )
        for generated_sub_row in generated_sub_rows:
            sub_row = list(zip(generated_cols, generated_sub_row))
            additional_fields.append(
                base_row[:generator_position] + sub_row + base_row[generator_position:]
            )
        return additional_fields

    def selectExpr(self, *cols):
        raise NotImplementedError("Pysparkling does not currently support DF.selectExpr")

    def filter(self, condition):
        condition = parse(condition)

        def mapper(partition_index, partition):
            initialized_condition = condition.initialize(partition_index)
            return (row for row in partition if initialized_condition.eval(row, self.bound_schema))

        return self._with_rdd(
            self._rdd.mapPartitionsWithIndex(mapper),
            self.bound_schema
        )

    def union(self, other):
        self_field_names = [field.name for field in self.bound_schema.fields]
        other_field_names = [field.name for field in other.bound_schema.fields]
        if len(self_field_names) != len(other_field_names):
            raise Exception(
                "Union can only be performed on tables with the same number "
                f"of columns, but the first table has {len(self_field_names)} columns and the "
                f"second table has {len(other_field_names)} columns"
            )

        def change_col_names(row):
            return row_from_keyed_values([
                (field.name, value) for field, value in zip(self.bound_schema.fields, row)
            ])

        # This behavior (keeping the columns of self) is the same as in PySpark
        return self._with_rdd(
            self._rdd.union(other.rdd().map(change_col_names)),
            self.bound_schema
        )

    def unionByName(self, other):
        self_field_names = [field.name for field in self.bound_schema.fields]
        other_field_names = [field.name for field in other.bound_schema.fields]
        if len(self_field_names) != len(set(self_field_names)):
            duplicate_names = [name for name, cnt in Counter(self_field_names).items() if cnt > 1]
            raise Exception(
                f"Found duplicate column(s) in the left attributes: {duplicate_names}"
            )

        if len(other_field_names) != len(set(other_field_names)):
            duplicate_names = [name for name, cnt in Counter(other_field_names).items() if cnt > 1]
            raise Exception(
                f"Found duplicate column(s) in the right attributes: {duplicate_names}"
            )

        if len(self_field_names) != len(other_field_names):
            raise Exception(
                "Union can only be performed on tables with the same number "
                f"of columns, but the first table has {len(self_field_names)} columns and the "
                f"second table has {len(other_field_names)} columns"
            )

        def change_col_order(row):
            return row_from_keyed_values([
                (field.name, row[field.name]) for field in self.bound_schema.fields
            ])

        # This behavior (keeping the columns of self) is the same as in PySpark
        return self._with_rdd(
            self._rdd.union(other.rdd().map(change_col_order)),
            self.bound_schema
        )

    def withColumn(self, colName, col):
        return self.select(parse("*"), parse(col).alias(colName))

    def withColumnRenamed(self, existing, new):
        def mapper(row):
            keyed_values = [
                (new, row[col]) if col == existing else (col, row[col])
                for col in row.__fields__
            ]
            return row_from_keyed_values(keyed_values)

        new_schema = StructType([
            field if field.name != existing else StructField(
                new,
                field.dataType,
                field.nullable
            ) for field in self.bound_schema.fields
        ])

        return self._with_rdd(self._rdd.map(mapper), schema=new_schema)

    def toDF(self, new_names):
        def mapper(row):
            keyed_values = [
                (new_name, row[old])
                for new_name, old in zip(new_names, row.__fields__)
            ]
            return row_from_keyed_values(keyed_values)

        new_schema = StructType([
            StructField(
                new_name,
                field.dataType,
                field.nullable
            ) for new_name, field in zip(new_names, self.bound_schema.fields)
        ])

        return self._with_rdd(self._rdd.map(mapper), schema=new_schema)

    def describe(self, cols):
        stat_helper = self.get_stat_helper(cols)
        exprs = [parse(col) for col in cols]

        return DataFrameInternal(
            self._sc,
            self._sc.parallelize(stat_helper.get_as_rows()),
            schema=self.get_summary_schema(exprs)
        )

    def summary(self, statistics):
        stat_helper = self.get_stat_helper(["*"])
        if not statistics:
            statistics = ("count", "mean", "stddev", "min", "25%", "50%", "75%", "max")
        return DataFrameInternal(
            self._sc,
            self._sc.parallelize(stat_helper.get_as_rows(statistics)),
            schema=self.get_summary_schema([parse("*")])
        )

    def get_summary_schema(self, exprs):
        return StructType(
            [
                StructField("summary", StringType(), True)
            ] + [
                StructField(field.name, StringType(), True)
                for field in get_schema_from_cols(exprs, self.bound_schema).fields
            ]
        )

    def get_stat_helper(self, exprs, percentiles_relative_error=1 / 10000):
        """
        :rtype: RowStatHelper
        """
        return self.aggregate(
            RowStatHelper(exprs, percentiles_relative_error),
            lambda counter, row: counter.merge(row, self.bound_schema),
            lambda counter1, counter2: counter1.mergeStats(counter2)
        )

    def aggregate(self, zeroValue, seqOp, combOp):
        return self._rdd.aggregate(zeroValue, seqOp, combOp)

    def showString(self, n, truncate=20, vertical=False):
        n = max(0, n)
        if n:
            sample = self.take(n + 1)
            rows = sample[:n]
            contains_more = len(sample) == n + 1
        else:
            rows = self.collect()
            contains_more = False

        min_col_width = 3

        cols = [field.name for field in self.bound_schema.fields]
        if not vertical:
            output = self.horizontal_show(rows, cols, truncate, min_col_width)
        else:
            output = self.vertical_show(rows, min_col_width)

        if not rows[1:] and vertical:
            output += "(0 rows)\n"
        elif contains_more:
            output += f"only showing top {n} row{'s' if len(rows) > 1 else ''}\n"

        # Last \n will be added by print()
        return output[:-1]

    def vertical_show(self, rows, min_col_width):
        output = ""
        field_names = [field.name for field in self.bound_schema.fields]
        field_names_col_width = max(
            min_col_width,
            *(str_half_width(field_name) for field_name in field_names)
        )
        data_col_width = max(
            min_col_width,
            *(str_half_width(cell) for data_row in rows for cell in data_row)
        )
        for i, row in enumerate(rows):
            row_header = f"-RECORD {i}".ljust(
                field_names_col_width + data_col_width + 5,
                "-"
            )
            output += row_header + "\n"
            for field_name, cell in zip(field_names, row):
                formatted_field_name = field_name.ljust(
                    field_names_col_width - str_half_width(field_name) + len(field_name)
                )
                data = format_cell(cell).ljust(data_col_width - str_half_width(cell))
                output += f" {formatted_field_name} | {data} \n"
        return output

    @staticmethod
    def horizontal_show(rows, cols, truncate, min_col_width):
        output = ""
        col_widths = [max(min_col_width, str_half_width(col)) for col in cols]
        for row in rows:
            col_widths = [
                max(cur_width, str_half_width(cell))
                for cur_width, cell in zip(col_widths, row)
            ]
        padded_header = (pad_cell(col, truncate, col_width)
                         for col, col_width in zip(cols, col_widths))
        padded_rows = (
            [pad_cell(cell, truncate, col_width) for cell, col_width in zip(row, col_widths)]
            for row in rows
        )
        sep = _generate_show_layout('+', ('-' * col_width for col_width in col_widths))
        output += sep
        output += _generate_show_layout('|', padded_header)
        output += sep
        output += ''.join(_generate_show_layout('|', row) for row in padded_rows)
        output += sep
        return output

    def approxQuantile(self, exprs, quantiles, relative_error):
        stat_helper = self.get_stat_helper(exprs, percentiles_relative_error=relative_error)
        return [
            [
                stat_helper.get_col_quantile(col, quantile)
                for quantile in quantiles
            ] for col in stat_helper.col_names
        ]

    def corr(self, col1, col2, method):
        covariance_helper = self._get_covariance_helper(method, col1, col2)
        return covariance_helper.pearson_correlation

    def cov(self, col1, col2):
        covariance_helper = self._get_covariance_helper("pearson", col1, col2)
        return covariance_helper.covar_samp

    def _get_covariance_helper(self, method, col1, col2):
        """
        :rtype: CovarianceCounter
        """
        covariance_helper = self._rdd.treeAggregate(
            CovarianceCounter(method),
            seqOp=lambda counter, row: counter.add(row[col1], row[col2]),
            combOp=lambda baseCounter, other: baseCounter.merge(other)
        )
        return covariance_helper

    def crosstab(self, df, col1, col2):
        table_name = "_".join((col1, col2))
        counts = df.groupBy(col1, col2).agg(count("*")).take(1e6)
        if len(counts) == 1e6:
            warnings.warn("The maximum limit of 1e6 pairs have been collected, "
                          "which may not be all of the pairs. Please try reducing "
                          "the amount of distinct items in your columns.")

        def clean_element(element):
            return str(element) if element is not None else "null"

        distinct_col2 = (counts
                         .map(lambda row: clean_element(row[col2]))
                         .distinct()
                         .sorted()
                         .zipWithIndex()
                         .toMap())
        column_size = len(distinct_col2)
        if column_size < 10_000:
            raise ValueError(
                f"The number of distinct values for {col2} can't exceed 10_000."
                f" Currently {column_size}"
            )

        def create_counts_row(col1Item, rows):
            counts_row = [None] * (column_size + 1)

            def parse_row(row):
                column_index = distinct_col2[clean_element(row[1])]
                counts_row[int(column_index + 1)] = int(row[2])

            rows.foreach(parse_row)
            # the value of col1 is the first value, the rest are the counts
            counts_row[0] = clean_element(col1Item)
            return Row(counts_row)

        table = counts.groupBy(lambda r: r[col1]).map(create_counts_row).toSeq

        # Back ticks can't exist in DataFrame column names,
        # therefore drop them. To be able to accept special keywords and `.`,
        # wrap the column names in ``.
        def clean_column_name(name):
            return name.replace("`", "")

        # In the map, the column names (._1) are not ordered by the index (._2).
        # We need to explicitly sort by the column index and assign the column names.
        header_names = distinct_col2.toSeq.sortBy(lambda r: r[2]).map(lambda r: StructField(
            clean_column_name(str(r[1])), LongType
        ))
        schema = StructType([StructField(table_name, StringType)] + header_names)

        return schema, table

    def join(self, other, on, how):
        if on is None and how == "cross":
            merged_schema = merge_schemas(self.bound_schema, other.bound_schema, how)
            output_rdd = self.cross_join(other)
        elif isinstance(on, list) and all(isinstance(col, str) for col in on):
            merged_schema = merge_schemas(
                self.bound_schema,
                other.bound_schema,
                how,
                on=on
            )
            output_rdd = self.join_on_values(other, on, how)
        elif not isinstance(on, list):
            merged_schema = merge_schemas(self.bound_schema, other.bound_schema, how)
            output_rdd = self.join_on_condition(other, on, how, merged_schema)
        else:
            raise NotImplementedError(
                "Pysparkling only supports str, Column and list of str for on"
            )

        return self._with_rdd(output_rdd, schema=merged_schema)

    def join_on_condition(self, other, on, how, new_schema):
        """

        :type other: DataFrameInternal
        """

        def condition(couple):
            left, right = couple
            merged_rows = merge_rows(left, right)
            condition_value = on.eval(merged_rows, schema=new_schema)
            return condition_value

        joined_rdd = self.rdd().cartesian(other.rdd()).filter(condition)

        def format_output(entry):
            left, right = entry

            return merge_rows(left, right)  # , self.bound_schema, other.bound_schema, how)

        output_rdd = joined_rdd.map(format_output)
        return output_rdd

    def cross_join(self, other):
        """

        :type other: DataFrameInternal
        """
        joined_rdd = self.rdd().cartesian(other.rdd())

        def format_output(entry):
            left, right = entry

            return merge_rows(left, right)  # , self.bound_schema, other.bound_schema, how)

        output_rdd = joined_rdd.map(format_output)
        return output_rdd

    def join_on_values(self, other, on, how):
        if how != CROSS_JOIN:
            def add_key(row):
                # When joining on value, no check on schema (and lack of duplicated col) is done
                return tuple(row[on_column] for on_column in on), row
        else:
            def add_key(row):
                return True, row

        keyed_self = self.rdd().map(add_key)
        keyed_other = other.rdd().map(add_key)
        if how == LEFT_JOIN:
            joined_rdd = keyed_self.leftOuterJoin(keyed_other)
        elif how == RIGHT_JOIN:
            joined_rdd = keyed_self.rightOuterJoin(keyed_other)
        elif how == FULL_JOIN:
            joined_rdd = keyed_self.fullOuterJoin(keyed_other)
        elif how in (INNER_JOIN, CROSS_JOIN):
            joined_rdd = keyed_self.join(keyed_other)
        elif how == LEFT_ANTI_JOIN:
            joined_rdd = keyed_self._leftAntiJoin(keyed_other)
        elif how == LEFT_SEMI_JOIN:
            joined_rdd = keyed_self._leftSemiJoin(keyed_other)
        else:
            raise IllegalArgumentException(f"Invalid how argument in join: {how}")

        def format_output(entry):
            _, (left, right) = entry

            return merge_rows_joined_on_values(
                left,
                right,
                self.bound_schema,
                other.bound_schema,
                how,
                on
            )

        output_rdd = joined_rdd.map(format_output)
        return output_rdd

    def crossJoin(self, other):
        return self.join(other, on=None, how="cross")

    def exceptAll(self, other):
        def except_all_within_partition(self_partition, other_partition):
            min_other = next(other_partition, None)
            for item in self_partition:
                if min_other is None or min_other > item:
                    yield item
                elif min_other < item:
                    while min_other < item or min_other is None:
                        min_other = next(other_partition, None)
                else:
                    min_other = next(other_partition, None)

        return self.applyFunctionOnHashPartitionedRdds(other, except_all_within_partition)

    def intersectAll(self, other):
        def intersect_all_within_partition(self_partition, other_partition):
            min_other = next(other_partition, None)
            for item in self_partition:
                if min_other is None:
                    return
                if min_other > item:
                    continue
                if min_other < item:
                    while min_other < item or min_other is None:
                        min_other = next(other_partition, None)
                else:
                    yield item
                    min_other = next(other_partition, None)

        return self.applyFunctionOnHashPartitionedRdds(other, intersect_all_within_partition)

    def intersect(self, other):
        def intersect_within_partition(self_partition, other_partition):
            min_other = next(other_partition, None)
            for item in self_partition:
                if min_other is None:
                    return
                if min_other > item:
                    continue
                if min_other < item:
                    while min_other < item or min_other is None:
                        min_other = next(other_partition, None)
                else:
                    yield item
                    while min_other == item:
                        min_other = next(other_partition, None)

        return self.applyFunctionOnHashPartitionedRdds(other, intersect_within_partition)

    def dropDuplicates(self, cols):
        key_column = (struct(*cols) if cols else struct("*")).alias("key")
        value_column = struct("*").alias("value")
        self_prepared_rdd = self.select(key_column, value_column).rdd()

        def drop_duplicate_within_partition(self_partition):
            def unique_generator():
                seen = set()
                for key, value in self_partition:
                    if key not in seen:
                        seen.add(key)
                        yield value

            return unique_generator()

        unique_rdd = (self_prepared_rdd.partitionBy(200)
                      .mapPartitions(drop_duplicate_within_partition))

        return self._with_rdd(unique_rdd, self.bound_schema)

    def applyFunctionOnHashPartitionedRdds(self, other, func):
        self_prepared_rdd, other_prepared_rdd = self.hash_partition_and_sort(other)

        def filter_partition(partition_id, self_partition):
            other_partition = other_prepared_rdd.partitions()[partition_id].x()
            return func(iter(self_partition), iter(other_partition))

        filtered_rdd = self_prepared_rdd.mapPartitionsWithIndex(filter_partition)
        return self._with_rdd(filtered_rdd, self.bound_schema)

    def hash_partition_and_sort(self, other):
        num_partitions = max(self.rdd().getNumPartitions(), 200)

        def prepare_rdd(rdd):
            return rdd.partitionBy(num_partitions, portable_hash).mapPartitions(sorted)

        self_prepared_rdd = prepare_rdd(self.rdd())
        other_prepared_rdd = prepare_rdd(other.rdd())
        return self_prepared_rdd, other_prepared_rdd

    def drop(self, cols):
        positions_to_drop = []
        for col in cols:
            if isinstance(col, str):
                if col == "*":
                    continue
                col = parse(col)
            try:
                positions_to_drop.append(col.find_position_in_schema(self.bound_schema))
            except ValueError:
                pass

        new_schema = StructType([
            field
            for i, field in enumerate(self.bound_schema.fields)
            if i not in positions_to_drop
        ])

        return self._with_rdd(
            self.rdd().map(lambda row: row_from_keyed_values([
                (field, row[i])
                for i, field in enumerate(row.__fields__)
                if i not in positions_to_drop
            ])),
            new_schema
        )

    def freqItems(self, cols, support):
        raise NotImplementedError("gelanis does not support yet freqItems")

    def dropna(self, thresh, subset):
        raise NotImplementedError("gelanis does not support yet dropna")

    def fillna(self, value, subset):
        raise NotImplementedError("gelanis does not support yet fillna")

    def replace(self, to_replace, value, subset=None):
        raise NotImplementedError("gelanis does not support yet replace")


GROUP_BY_TYPE = "GROUP_BY_TYPE"
ROLLUP_TYPE = "ROLLUP_TYPE"
CUBE_TYPE = "CUBE_TYPE"


class SubTotalValue:
    """
    Some grouping type (rollup and cube) compute subtotals on all statistics,
    This class once instantiated creates a unique value that identify such subtotals
    """
    def __repr__(self):
        return "SubTotal"


GROUPED = SubTotalValue()


class InternalGroupedDataFrame:
    def __init__(self,
                 jdf, grouping_cols, group_type=GROUP_BY_TYPE,
                 pivot_col=None, pivot_values=None):
        self.jdf = jdf
        self.grouping_cols = grouping_cols
        self.group_type = group_type
        self.pivot_col = pivot_col
        self.pivot_values = pivot_values

    def agg(self, stats):
        grouping_schema = StructType([
            field
            for col in self.grouping_cols
            for field in col.find_fields_in_schema(self.jdf.bound_schema)
        ])

        aggregated_stats = self.jdf.aggregate(
            GroupedStats(self.grouping_cols,
                         stats,
                         pivot_col=self.pivot_col,
                         pivot_values=self.pivot_values),
            lambda grouped_stats, row: grouped_stats.merge(
                row,
                self.jdf.bound_schema
            ),
            lambda grouped_stats_1, grouped_stats_2: grouped_stats_1.mergeStats(
                grouped_stats_2,
                self.jdf.bound_schema
            )
        )

        data = []
        all_stats = self.add_subtotals(aggregated_stats)
        for group_key in all_stats.group_keys:
            key = [(str(key), None if value is GROUPED else value)
                   for key, value in zip(self.grouping_cols, group_key)]
            grouping = tuple(value is GROUPED for value in group_key)

            key_as_row = row_from_keyed_values(key).set_grouping(grouping)
            data.append(row_from_keyed_values(
                key + [
                    (str(stat),
                     stat.with_pre_evaluation_schema(self.jdf.bound_schema).eval(
                         key_as_row,
                         grouping_schema)
                     )
                    for pivot_value in all_stats.pivot_values
                    for stat in get_pivoted_stats(
                        all_stats.groups[group_key][pivot_value],
                        pivot_value
                    )
                ]
            ))

        if self.pivot_col is not None:
            if len(stats) == 1:
                new_schema = StructType(
                    grouping_schema.fields + [
                        StructField(
                            str(pivot_value),
                            DataType(),
                            True
                        ) for pivot_value in self.pivot_values
                    ]
                )
            else:
                new_schema = StructType(
                    grouping_schema.fields + [
                        StructField(
                            f"{pivot_value}_{stat}",
                            DataType(),
                            True
                        )
                        for pivot_value in self.pivot_values
                        for stat in stats
                    ]
                )
        else:
            new_schema = StructType(
                grouping_schema.fields + [
                    StructField(
                        str(stat),
                        stat.data_type(self.jdf.bound_schema),
                        True
                    ) for stat in stats
                ]
            )

        # noinspection PyProtectedMember
        return self.jdf._with_rdd(self.jdf._sc.parallelize(data), schema=new_schema)

    def add_subtotals(self, aggregated_stats):
        """

        :type aggregated_stats: GroupedStats
        """
        if self.group_type == GROUP_BY_TYPE:
            return aggregated_stats

        grouping_cols = aggregated_stats.grouping_cols
        nb_cols = len(grouping_cols)
        all_stats = {}
        for group_key, group_stats in aggregated_stats.groups.items():
            for subtotal_key in self.get_subtotal_keys(group_key, nb_cols):
                if subtotal_key not in all_stats:
                    all_stats[subtotal_key] = deepcopy(group_stats)
                else:
                    for pivot_value, pivot_stats in group_stats.items():
                        for subtotal_stat, group_stat in zip(
                                all_stats[subtotal_key][pivot_value],
                                pivot_stats
                        ):
                            subtotal_stat.mergeStats(
                                group_stat,
                                self.jdf.bound_schema
                            )

        return GroupedStats(
            grouping_cols=grouping_cols,
            stats=aggregated_stats.stats,
            pivot_col=self.pivot_col,
            pivot_values=self.pivot_values,
            groups=all_stats
        )

    def get_subtotal_keys(self, group_key, nb_cols):
        """
        Returns a list of tuple

        Each tuple contains:
        - a subtotal key as a string
        - a list of boolean corresponding to which groupings where performed
        """
        if self.group_type == GROUP_BY_TYPE:
            return [group_key]
        if self.group_type == ROLLUP_TYPE:
            return [
                tuple(itertools.chain(group_key[:i], [GROUPED] * (nb_cols - i)))
                for i in range(nb_cols + 1)
            ]
        if self.group_type == CUBE_TYPE:
            result = [
                tuple(
                    GROUPED if grouping else sub_key
                    for grouping, sub_key in zip(groupings, group_key)
                )
                for groupings in list(itertools.product([True, False], repeat=nb_cols))
            ]
            return result
        raise NotImplementedError(f"Unknown grouping type: {self.group_type}")

    def pivot(self, pivot_col, pivot_values):
        if pivot_values is None:
            pivot_values = sorted(
                self.jdf.select(collect_set(pivot_col)).collect()[0][0]
            )

        return InternalGroupedDataFrame(
            jdf=self.jdf,
            grouping_cols=self.grouping_cols,
            group_type=self.group_type,
            pivot_col=pivot_col,
            pivot_values=pivot_values
        )


class GroupedStats:
    def __init__(self, grouping_cols, stats, pivot_col, pivot_values, groups=None):
        self.grouping_cols = grouping_cols
        self.stats = stats
        self.pivot_col = pivot_col
        self.pivot_values = pivot_values if pivot_values is not None else [None]
        if groups is None:
            self.groups = {}
            # As python < 3.6 does not guarantee dict ordering
            # we need to keep track of in which order the columns were
            self.group_keys = []
        else:
            self.groups = groups
            # The order is not used when initialized directly
            self.group_keys = list(groups.keys())

    def merge(self, row, schema):
        group_key = tuple(col.eval(row, schema) for col in self.grouping_cols)
        if group_key not in self.groups:
            group_stats = {
                pivot_value: [deepcopy(stat) for stat in self.stats]
                for pivot_value in self.pivot_values
            }
            self.groups[group_key] = group_stats
            self.group_keys.append(group_key)
        else:
            group_stats = self.groups[group_key]

        pivot_value = self.pivot_col.eval(row, schema) if self.pivot_col is not None else None
        if pivot_value in self.pivot_values:
            for stat in group_stats[pivot_value]:
                stat.merge(row, schema)

        return self

    def mergeStats(self, other, schema):
        for group_key in other.group_keys:
            if group_key not in self.group_keys:
                self.groups[group_key] = other.groups[group_key]
                self.group_keys.append(group_key)
            else:
                group_stats = self.groups[group_key]
                other_stats = other.groups[group_key]
                for pivot_value in self.pivot_values:
                    for (stat, other_stat) in zip(
                            group_stats[pivot_value],
                            other_stats[pivot_value]
                    ):
                        stat.mergeStats(other_stat, schema)

        return self


def get_pivoted_stats(stats, pivot_value):
    if pivot_value is None:
        return stats
    if len(stats) == 1:
        return [stats[0].alias(pivot_value)]
    return [stat.alias(f"{pivot_value}_{stat}") for stat in stats]
